"""
Produce integration test data, which is tested by the `test_integration_features.py`
tests. One thing to note here is that all redshifts are reasonably high.

This is necessary, because low redshifts mean that neutral fractions are small,
and then numerical noise gets relatively more important, and can make the comparison
fail at the tens-of-percent level.
"""
import click
import glob
import h5py
import logging
import numpy as np
import os
import questionary as qs
import sys
import tempfile
from pathlib import Path
from powerbox import get_power

from py21cmfast import (
    AstroParams,
    CosmoParams,
    FlagOptions,
    InitialConditions,
    UserParams,
    config,
    determine_halo_list,
    global_params,
    initial_conditions,
    perturb_field,
    perturb_halo_list,
    run_coeval,
    run_lightcone,
)

logger = logging.getLogger("py21cmfast")
logging.basicConfig()

SEED = 12345
DATA_PATH = Path(__file__).parent / "test_data"
DEFAULT_USER_PARAMS = {
    "HII_DIM": 50,
    "DIM": 150,
    "BOX_LEN": 100,
    "NO_RNG": True,
    "USE_INTERPOLATION_TABLES": True,
}
DEFAULT_ZPRIME_STEP_FACTOR = 1.04

LIGHTCONE_FIELDS = [
    "density",
    "velocity",
    "xH_box",
    "Ts_box",
    "z_re_box",
    "Gamma12_box",
    "dNrec_box",
    "x_e_box",
    "Tk_box",
    "J_21_LW_box",
    "brightness_temp",
]

COEVAL_FIELDS = [
    "lowres_density",
    "lowres_vx_2LPT",
    "lowres_vx",
] + LIGHTCONE_FIELDS

OPTIONS = {
    "simple": [12, {}],
    "perturb_high_res": [12, {"PERTURB_ON_HIGH_RES": True}],
    "change_step_factor": [11, {"zprime_step_factor": 1.02}],
    "change_z_heat_max": [30, {"z_heat_max": 40}],
    "larger_step_factor": [
        13,
        {"zprime_step_factor": 1.05, "z_heat_max": 25, "HMF": 0},
    ],
    "interp_perturb_field": [16, {"interp_perturb_field": True}],
    "mdzeta": [14, {"USE_MASS_DEPENDENT_ZETA": True}],
    "rsd": [9, {"SUBCELL_RSD": True}],
    "inhomo": [10, {"INHOMO_RECO": True}],
    "tsfluct": [16, {"HMF": 3, "USE_TS_FLUCT": True}],
    "mmin_in_mass": [20, {"z_heat_max": 45, "M_MIN_in_Mass": True, "HMF": 2}],
    "fftw_wisdom": [35, {"USE_FFTW_WISDOM": True}],
    "mini_halos": [
        18,
        {
            "z_heat_max": 25,
            "USE_MINI_HALOS": True,
            "USE_MASS_DEPENDENT_ZETA": True,
            "INHOMO_RECO": True,
            "USE_TS_FLUCT": True,
            "zprime_step_factor": 1.1,
            "N_THREADS": 4,
            "USE_FFTW_WISDOM": True,
            "NUM_FILTER_STEPS_FOR_Ts": 8,
        },
    ],
    "nthreads": [8, {"N_THREADS": 2}],
    "photoncons": [10, {"PHOTON_CONS": True}],
    "mdz_and_photoncons": [
        8.5,
        {
            "USE_MASS_DEPENDENT_ZETA": True,
            "PHOTON_CONS": True,
            "z_heat_max": 25,
            "zprime_step_factor": 1.1,
        },
    ],
    "mdz_and_ts_fluct": [
        9,
        {
            "USE_MASS_DEPENDENT_ZETA": True,
            "USE_TS_FLUCT": True,
            "INHOMO_RECO": True,
            "PHOTON_CONS": True,
            "z_heat_max": 25,
            "zprime_step_factor": 1.1,
        },
    ],
    "minimize_mem": [
        9,
        {
            "USE_MASS_DEPENDENT_ZETA": True,
            "USE_TS_FLUCT": True,
            "INHOMO_RECO": True,
            "PHOTON_CONS": True,
            "z_heat_max": 25,
            "zprime_step_factor": 1.1,
            "MINIMIZE_MEMORY": True,
        },
    ],
    "mdz_and_tsfluct_nthreads": [
        8.5,
        {
            "N_THREADS": 2,
            "USE_FFTW_WISDOM": True,
            "USE_MASS_DEPENDENT_ZETA": True,
            "INHOMO_RECO": True,
            "USE_TS_FLUCT": True,
            "PHOTON_CONS": True,
            "z_heat_max": 25,
            "zprime_step_factor": 1.1,
        },
    ],
    "halo_field": [9, {"USE_HALO_FIELD": True}],
    "halo_field_mdz": [
        8.5,
        {
            "USE_MASS_DEPENDENT_ZETA": True,
            "USE_HALO_FIELD": True,
            "USE_TS_FLUCT": True,
            "z_heat_max": 25,
            "zprime_step_factor": 1.1,
        },
    ],
    "halo_field_mdz_highres": [
        8.5,
        {
            "USE_MASS_DEPENDENT_ZETA": True,
            "USE_HALO_FIELD": True,
            "USE_TS_FLUCT": False,
            "PERTURB_ON_HIGH_RES": True,
            "N_THREADS": 4,
            "z_heat_max": 25,
            "zprime_step_factor": 1.1,
        },
    ],
    "mdz_tsfluct_nthreads": [
        12.0,
        {
            "USE_MASS_DEPENDENT_ZETA": True,
            "USE_TS_FLUCT": True,
            "PERTURB_ON_HIGH_RES": False,
            "N_THREADS": 4,
            "z_heat_max": 25,
            "zprime_step_factor": 1.2,
            "NUM_FILTER_STEPS_FOR_Ts": 4,
            "USE_INTERPOLATION_TABLES": False,
        },
    ],
    "ts_fluct_no_tables": [
        12.0,
        {
            "USE_TS_FLUCT": True,
            "N_THREADS": 4,
            "z_heat_max": 25,
            "zprime_step_factor": 1.2,
            "NUM_FILTER_STEPS_FOR_Ts": 4,
            "USE_INTERPOLATION_TABLES": False,
        },
    ],
    "minihalos_no_tables": [
        12.0,
        {
            "USE_MINI_HALOS": True,
            "USE_MASS_DEPENDENT_ZETA": True,
            "USE_TS_FLUCT": True,
            "N_THREADS": 4,
            "z_heat_max": 25,
            "zprime_step_factor": 1.1,
            "NUM_FILTER_STEPS_FOR_Ts": 4,
            "USE_INTERPOLATION_TABLES": False,
        },
    ],
    "fast_fcoll_hiz": [
        18,
        {"N_THREADS": 4, "FAST_FCOLL_TABLES": True, "USE_INTERPOLATION_TABLES": True},
    ],
    "fast_fcoll_lowz": [
        8,
        {"N_THREADS": 4, "FAST_FCOLL_TABLES": True, "USE_INTERPOLATION_TABLES": True},
    ],
    "relvel": [
        18,
        {
            "z_heat_max": 25,
            "USE_MINI_HALOS": True,
            "zprime_step_factor": 1.1,
            "N_THREADS": 4,
            "NUM_FILTER_STEPS_FOR_Ts": 8,
            "USE_INTERPOLATION_TABLES": True,
            "FAST_FCOLL_TABLES": True,
            "USE_RELATIVE_VELOCITIES": True,
        },
    ],
}

if len(set(OPTIONS.keys())) != len(list(OPTIONS.keys())):
    raise ValueError("There is a non-unique option name!")

OPTIONS_PT = {
    "simple": [10, {}],
    "no2lpt": [10, {"USE_2LPT": False}],
    "linear": [10, {"EVOLVE_DENSITY_LINEARLY": 1}],
    "highres": [10, {"PERTURB_ON_HIGH_RES": True}],
}

if len(set(OPTIONS_PT.keys())) != len(list(OPTIONS_PT.keys())):
    raise ValueError("There is a non-unique option_pt name!")

OPTIONS_HALO = {"halo_field": [9, {"USE_HALO_FIELD": True}]}

if len(set(OPTIONS_HALO.keys())) != len(list(OPTIONS_HALO.keys())):
    raise ValueError("There is a non-unique option_halo name!")


def get_defaults(kwargs, cls):
    return {k: kwargs.get(k, v) for k, v in cls._defaults_.items()}


def get_all_defaults(kwargs):
    flag_options = get_defaults(kwargs, FlagOptions)
    astro_params = get_defaults(kwargs, AstroParams)
    cosmo_params = get_defaults(kwargs, CosmoParams)
    user_params = get_defaults(kwargs, UserParams)
    return user_params, cosmo_params, astro_params, flag_options


def get_all_options(redshift, **kwargs):
    user_params, cosmo_params, astro_params, flag_options = get_all_defaults(kwargs)
    user_params.update(DEFAULT_USER_PARAMS)
    out = {
        "redshift": redshift,
        "user_params": user_params,
        "cosmo_params": cosmo_params,
        "astro_params": astro_params,
        "flag_options": flag_options,
        "use_interp_perturb_field": kwargs.get("use_interp_perturb_field", False),
        "random_seed": SEED,
    }

    for key in kwargs:
        if key.upper() in (k.upper() for k in global_params.keys()):
            out[key] = kwargs[key]
    return out


def get_all_options_ics(**kwargs):
    user_params, cosmo_params, astro_params, flag_options = get_all_defaults(kwargs)
    user_params.update(DEFAULT_USER_PARAMS)
    out = {
        "user_params": user_params,
        "cosmo_params": cosmo_params,
        "random_seed": SEED,
    }

    for key in kwargs:
        if key.upper() in (k.upper() for k in global_params.keys()):
            out[key] = kwargs[key]
    return out


def get_all_options_halo(redshift, **kwargs):
    user_params, cosmo_params, astro_params, flag_options = get_all_defaults(kwargs)
    user_params.update(DEFAULT_USER_PARAMS)
    out = {
        "redshift": redshift,
        "user_params": user_params,
        "cosmo_params": cosmo_params,
        "astro_params": astro_params,
        "flag_options": flag_options,
        "random_seed": SEED,
    }

    for key in kwargs:
        if key.upper() in (k.upper() for k in global_params.keys()):
            out[key] = kwargs[key]
    return out


def produce_coeval_power_spectra(redshift, **kwargs):
    options = get_all_options(redshift, **kwargs)

    coeval = run_coeval(write=write_ics_only_hook, **options)
    p = {}

    for field in COEVAL_FIELDS:
        if hasattr(coeval, field):
            p[field], k = get_power(
                getattr(coeval, field), boxlength=coeval.user_params.BOX_LEN
            )

    return k, p, coeval


def produce_lc_power_spectra(redshift, **kwargs):
    options = get_all_options(redshift, **kwargs)
    lightcone = run_lightcone(
        max_redshift=options["redshift"] + 2,
        lightcone_quantities=[
            k
            for k in LIGHTCONE_FIELDS
            if (
                options["flag_options"].get("USE_TS_FLUCT", False)
                or k not in ("Ts_box", "x_e_box", "Tk_box", "J_21_LW_box")
            )
        ],
        write=write_ics_only_hook,
        **options,
    )

    p = {}
    for field in LIGHTCONE_FIELDS:
        if hasattr(lightcone, field):
            p[field], k = get_power(
                getattr(lightcone, field), boxlength=lightcone.lightcone_dimensions
            )

    return k, p, lightcone


def produce_perturb_field_data(redshift, **kwargs):
    options = get_all_options(redshift, **kwargs)
    options_ics = get_all_options_ics(**kwargs)

    out = {
        key: kwargs[key]
        for key in kwargs
        if key.upper() in (k.upper() for k in global_params.keys())
    }

    velocity_normalisation = 1e16

    with config.use(regenerate=True, write=False):
        init_box = initial_conditions(**options_ics)
        pt_box = perturb_field(redshift=redshift, init_boxes=init_box, **out)

    p_dens, k_dens = get_power(
        pt_box.density,
        boxlength=options["user_params"]["BOX_LEN"],
    )
    p_vel, k_vel = get_power(
        pt_box.velocity * velocity_normalisation,
        boxlength=options["user_params"]["BOX_LEN"],
    )

    def hist(kind, xmin, xmax, nbins):
        data = getattr(pt_box, kind)
        if kind == "velocity":
            data = velocity_normalisation * data

        bins, edges = np.histogram(
            data,
            bins=np.linspace(xmin, xmax, nbins),
            range=[xmin, xmax],
            density=True,
        )

        left, right = edges[:-1], edges[1:]

        X = np.array([left, right]).T.flatten()
        Y = np.array([bins, bins]).T.flatten()
        return X, Y

    X_dens, Y_dens = hist("density", -0.8, 2.0, 50)
    X_vel, Y_vel = hist("velocity", -2, 2, 50)

    return k_dens, p_dens, k_vel, p_vel, X_dens, Y_dens, X_vel, Y_vel, init_box


def produce_halo_field_data(redshift, **kwargs):
    options_halo = get_all_options_halo(redshift, **kwargs)

    with config.use(regenerate=True, write=False):
        pt_halos = perturb_halo_list(**options_halo)

    return pt_halos


def get_filename(kind, name, **kwargs):
    # get sorted keys
    fname = f"{kind}_{name}.h5"
    return DATA_PATH / fname


def get_old_filename(redshift, kind, **kwargs):
    # get sorted keys
    kwargs = {k: kwargs[k] for k in sorted(kwargs)}
    string = "_".join(f"{k}={v}" for k, v in kwargs.items())
    fname = f"{kind}_z{redshift:.2f}_{string}.h5"

    return DATA_PATH / fname


def write_ics_only_hook(obj, **params):
    if isinstance(obj, InitialConditions):
        obj.write(**params)


def produce_power_spectra_for_tests(name, redshift, force, direc, **kwargs):
    fname = get_filename("power_spectra", name)

    # Need to manually remove it, otherwise h5py tries to add to it
    if fname.exists():
        if force:
            fname.unlink()
        else:
            print(f"Skipping {fname} because it already exists.")
            return fname

    # For tests, we *don't* want to use cached boxes, but we also want to use the
    # cache between the power spectra and lightcone. So we create a temporary
    # directory in which to cache results.
    with config.use(direc=direc):
        k, p, coeval = produce_coeval_power_spectra(redshift, **kwargs)
        k_l, p_l, lc = produce_lc_power_spectra(redshift, **kwargs)

    with h5py.File(fname, "w") as fl:
        for k, v in kwargs.items():
            fl.attrs[k] = v

        fl.attrs["HII_DIM"] = coeval.user_params.HII_DIM
        fl.attrs["DIM"] = coeval.user_params.DIM
        fl.attrs["BOX_LEN"] = coeval.user_params.BOX_LEN

        coeval_grp = fl.create_group("coeval")
        coeval_grp["k"] = k
        for key, val in p.items():
            coeval_grp[f"power_{key}"] = val

        lc_grp = fl.create_group("lightcone")
        lc_grp["k"] = k_l
        for key, val in p_l.items():
            lc_grp[f"power_{key}"] = val

        lc_grp["global_xH"] = lc.global_xH
        lc_grp["global_brightness_temp"] = lc.global_brightness_temp

    print(f"Produced {fname} with {kwargs}")
    return fname


def produce_data_for_perturb_field_tests(name, redshift, force, **kwargs):
    (
        k_dens,
        p_dens,
        k_vel,
        p_vel,
        X_dens,
        Y_dens,
        X_vel,
        Y_vel,
        init_box,
    ) = produce_perturb_field_data(redshift, **kwargs)

    fname = get_filename("perturb_field_data", name)

    # Need to manually remove it, otherwise h5py tries to add to it
    if os.path.exists(fname):
        if force:
            os.remove(fname)
        else:
            return fname

    with h5py.File(fname, "w") as fl:
        for k, v in kwargs.items():
            fl.attrs[k] = v

        fl.attrs["HII_DIM"] = init_box.user_params.HII_DIM
        fl.attrs["DIM"] = init_box.user_params.DIM
        fl.attrs["BOX_LEN"] = init_box.user_params.BOX_LEN

        fl["power_dens"] = p_dens
        fl["k_dens"] = k_dens

        fl["power_vel"] = p_vel
        fl["k_vel"] = k_vel

        fl["pdf_dens"] = Y_dens
        fl["x_dens"] = X_dens

        fl["pdf_vel"] = Y_vel
        fl["x_vel"] = X_vel

    print(f"Produced {fname} with {kwargs}")
    return fname


def produce_data_for_halo_field_tests(name, redshift, force, **kwargs):

    pt_halos = produce_halo_field_data(redshift, **kwargs)

    fname = get_filename("halo_field_data", name)

    # Need to manually remove it, otherwise h5py tries to add to it
    if os.path.exists(fname):
        if force:
            os.remove(fname)
        else:
            return fname

    with h5py.File(fname, "w") as fl:
        for k, v in kwargs.items():
            fl.attrs[k] = v

        fl["n_pt_halos"] = pt_halos.n_halos
        fl["pt_halo_masses"] = pt_halos.halo_masses

    print(f"Produced {fname} with {kwargs}")
    return fname


main = click.Group()


@main.command()
@click.option("--log-level", default="WARNING")
@click.option("--force/--no-force", default=False)
@click.option("--remove/--no-remove", default=True)
@click.option("--pt-only/--not-pt-only", default=False)
@click.option("--no-pt/--pt", default=False)
@click.option("--no-halo/--do-halo", default=False)
@click.option(
    "--names",
    multiple=True,
    type=click.Choice(list(OPTIONS.keys())),
    default=list(OPTIONS.keys()),
)
def go(
    log_level: str,
    force: bool,
    remove: bool,
    pt_only: bool,
    no_pt: bool,
    no_halo,
    names,
):
    logger.setLevel(log_level.upper())
    global_params.ZPRIME_STEP_FACTOR = DEFAULT_ZPRIME_STEP_FACTOR

    if names != list(OPTIONS.keys()):
        remove = False
        force = True

    if pt_only or no_pt or no_halo:
        remove = False

    # For tests, we *don't* want to use cached boxes, but we also want to use the
    # cache between the power spectra and lightcone. So we create a temporary
    # directory in which to cache results.
    direc = tempfile.mkdtemp()
    fnames = []

    if not pt_only:
        for name in names:
            redshift = OPTIONS[name][0]
            kwargs = OPTIONS[name][1]

            fnames.append(
                produce_power_spectra_for_tests(name, redshift, force, direc, **kwargs)
            )

    if not no_pt:
        for name, (redshift, kwargs) in OPTIONS_PT.items():
            fnames.append(
                produce_data_for_perturb_field_tests(name, redshift, force, **kwargs)
            )

    if not no_halo:
        for name, (redshift, kwargs) in OPTIONS_HALO.items():
            fnames.append(
                produce_data_for_halo_field_tests(name, redshift, force, **kwargs)
            )

    # Remove extra files that
    if not (names or pt_only or no_pt or no_halo):
        all_files = DATA_PATH.glob("*")
        for fl in all_files:
            if fl not in fnames:
                if remove:
                    print(f"Removing old file: {fl}")
                    os.remove(fl)
                else:
                    print(f"File is now redundant and can be removed: {fl}")


@main.command()
def convert():
    """Convert old-style data file names to new ones."""
    all_files = DATA_PATH.glob("*")

    old_names = {
        get_old_filename(v[0], "power_spectra", **v[1]): k for k, v in OPTIONS.items()
    }
    old_names_pt = {
        get_old_filename(v[0], "perturb_field_data", **v[1]): k
        for k, v in OPTIONS_PT.items()
    }
    old_names_hf = {
        get_old_filename(v[0], "halo_field_data", **v[1]): k
        for k, v in OPTIONS_HALO.items()
    }

    for fl in all_files:
        if fl.name.startswith("power_spectra"):
            if fl.stem.split("power_spectra_")[-1] in OPTIONS:
                continue
            elif fl in old_names:
                new_file = get_filename("power_spectra", old_names[fl])
                fl.rename(new_file)
                continue
        elif fl.name.startswith("perturb_field_data"):
            if fl.stem.split("perturb_field_data_")[-1] in OPTIONS_PT:
                continue
            elif fl in old_names_pt:
                new_file = get_filename("perturb_field_data", old_names_pt[fl])
                fl.rename(new_file)
                continue
        elif fl.name.startswith("halo_field_data"):
            if fl.stem.split("halo_field_data_")[-1] in OPTIONS_HALO:
                continue
            elif fl in old_names_hf:
                new_file = get_filename("halo_field_data", old_names_hf[fl])
                fl.rename(new_file)
                continue

        if qs.confirm(f"Remove {fl}?").ask():
            fl.unlink()


@main.command()
def clean():
    """Convert old-style data file names to new ones."""
    all_files = DATA_PATH.glob("*")

    for fl in all_files:
        if (
            fl.stem.startswith("power_spectra")
            and fl.stem.split("power_spectra_")[-1] in OPTIONS
        ):
            continue
        elif (
            fl.stem.startswith("perturb_field_data")
            and fl.stem.split("perturb_field_data_")[-1] in OPTIONS_PT
        ):
            continue
        elif (
            fl.stem.startswith("halo_field_data")
            and fl.stem.split("halo_field_data_")[-1] in OPTIONS_HALO
        ):
            continue

        if qs.confirm(f"Remove {fl}?").ask():
            fl.unlink()


if __name__ == "__main__":
    main()



"""
Tests for the tools in the wrapper.
"""

import pytest

import numpy as np

from py21cmfast import cache_tools


def test_query(ic):
    things = list(cache_tools.query_cache())

    print(things)

    classes = [t[1] for t in things]
    assert ic in classes


def test_bad_fname(tmpdirec):
    with pytest.raises(ValueError):
        cache_tools.readbox(direc=str(tmpdirec), fname="a_really_fake_file.h5")


def test_readbox_data(tmpdirec, ic):
    box = cache_tools.readbox(direc=str(tmpdirec), fname=ic.filename)

    assert np.all(box.hires_density == ic.hires_density)


def test_readbox_filter(ic, tmpdirec):
    ic2 = cache_tools.readbox(
        kind="InitialConditions", hsh=ic._md5, direc=str(tmpdirec)
    )
    assert np.all(ic2.hires_density == ic.hires_density)


def test_readbox_seed(ic, tmpdirec):
    ic2 = cache_tools.readbox(
        kind="InitialConditions",
        hsh=ic._md5,
        seed=ic.random_seed,
        direc=str(tmpdirec),
    )
    assert np.all(ic2.hires_density == ic.hires_density)


def test_readbox_nohash(ic, tmpdirec):
    with pytest.raises(ValueError):
        cache_tools.readbox(
            kind="InitialConditions", seed=ic.random_seed, direc=str(tmpdirec)
        )



# -*- coding: utf-8 -*-
"""Configuration options for the docs."""
from __future__ import unicode_literals

import os
import subprocess
import sys
from unittest.mock import MagicMock
from pathlib import Path

sys.path.insert(0, str(Path(__file__).absolute().parent.parent / "src"))


class Mock(MagicMock):
    """Make a Mock so that a package doesn't have to actually exist."""

    @classmethod
    def __getattr__(cls, name):
        """Get stuff."""
        return MagicMock()


MOCK_MODULES = [
    "py21cmfast.c_21cmfast",
    "click",
    "tqdm",
    "pyyaml",
    "scipy",
    "scipy.interpolate",
    "scipy.integrate",
    "scipy.special",
    "h5py",
    "cached_property",
]
sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)

# Get the version by running from shell -- this uses the git repo info, rather than
# requiring it to be installed.
out = subprocess.run(["python", "setup.py", "--version"], capture_output=True)

try:
    from py21cmfast import cache_tools
except ImportError:
    raise 
    
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.ifconfig",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.mathjax",
    "sphinx.ext.autosectionlabel",
    "numpydoc",
    "nbsphinx",
    "IPython.sphinxext.ipython_console_highlighting",
]
if os.getenv("SPELLCHECK"):
    extensions += ("sphinxcontrib.spelling",)
    spelling_show_suggestions = True
    spelling_lang = "en_US"

autosectionlabel_prefix_document = True

autosummary_generate = True
numpydoc_show_class_members = False

source_suffix = ".rst"
master_doc = "index"
project = "21cmFAST"
year = "2020"
author = "The 21cmFAST collaboration"
copyright = "{0}, {1}".format(year, author)
version = (
    release
) = out.stdout.decode().rstrip()  # Get it from `python setup.py --version`
templates_path = ["templates"]

pygments_style = "trac"
extlinks = {
    "issue": ("https://github.com/21cmFAST/21cmFAST/issues/%s", "#"),
    "pr": ("https://github.com/21cmFAST/21cmFAST/pull/%s", "PR #"),
}
# on_rtd is whether we are on readthedocs.org
on_rtd = os.environ.get("READTHEDOCS", None) == "True"

if not on_rtd:  # only set the theme if we're building docs locally
    html_theme = "sphinx_rtd_theme"

html_use_smartypants = True
html_last_updated_fmt = "%b %d, %Y"
html_split_index = False
html_sidebars = {"**": ["searchbox.html", "globaltoc.html", "sourcelink.html"]}
html_short_title = "%s-%s" % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False

mathjax_path = (
    "http://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_HTMLorMML"
)

exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "templates",
    "**.ipynb_checkpoints",
]



"""Run over output of logging to prepare for diffing."""
import sys

fname = sys.argv[1]

if len(sys.argv) > 2:
    REMOVE_LINE_NUMBERS = True
    print("REMOVING LINE NUMBERS ")
else:
    REMOVE_LINE_NUMBERS = False


with open(fname) as fl:
    lines = fl.readlines()
    out_lines = []
    for line in lines:
        bits = line.split("|")

        # Get rid of time.
        if len(bits) > 2:
            line = "|".join(bits[1:])

        # get rid of pid
        if len(bits) > 2:
            pid_bit = line.split("[pid=")[-1].split("]")[0]
            line = line.replace(pid_bit, "")

        if REMOVE_LINE_NUMBERS and len(bits) > 2:
            line_no = line.split(":")[1].split("[")[0].strip()
            line = line.replace(line_no, "")

        out_lines.append(line)

with open(fname + ".out", "w") as fl:
    fl.writelines(out_lines)



"""
Unit tests for output structures
"""

import pytest

import copy
import numpy as np
import pickle

from py21cmfast import InitialConditions  # An example of an output struct
from py21cmfast import IonizedBox, PerturbedField, TsBox, global_params


@pytest.fixture(scope="function")
def init(default_user_params):
    return InitialConditions(user_params=default_user_params)


@pytest.mark.parametrize("cls", [InitialConditions, PerturbedField, IonizedBox, TsBox])
def test_pointer_fields(cls):
    if cls is InitialConditions:
        inst = cls()
    else:
        with pytest.raises(KeyError):
            cls()

        inst = cls(redshift=7.0)

    # Get list of fields before and after array initialisation
    d = copy.copy(list(inst.__dict__.keys()))
    inst._init_arrays()
    new_names = [name for name in inst.__dict__ if name not in d]

    assert new_names
    assert all(n in inst.pointer_fields for n in new_names)


def test_non_existence(init, test_direc):
    assert not init.exists(direc=test_direc)


def test_writeability(init):
    """init is not initialized and therefore can't write yet."""
    with pytest.raises(IOError):
        init.write()


def test_readability(ic, tmpdirec, default_user_params):
    ic2 = InitialConditions(user_params=default_user_params)

    # without seeds, they are obviously exactly the same.
    assert ic._seedless_repr() == ic2._seedless_repr()

    assert ic2.exists(direc=tmpdirec)

    ic2.read(direc=tmpdirec)

    assert repr(ic) == repr(ic2)  # they should be exactly the same.
    assert str(ic) == str(ic2)  # their str is the same.
    assert hash(ic) == hash(ic2)
    assert ic == ic2
    assert ic is not ic2


def test_different_seeds(init, default_user_params):
    ic2 = InitialConditions(random_seed=2, user_params=default_user_params)

    assert init is not ic2
    assert init != ic2
    assert repr(init) != repr(ic2)
    assert init._seedless_repr() == ic2._seedless_repr()

    assert init._md5 == ic2._md5

    # make sure we didn't inadvertantly set the random seed while doing any of this
    assert init._random_seed is None


def test_pickleability(default_user_params):
    ic_ = InitialConditions(init=True, user_params=default_user_params)
    ic_.filled = True
    ic_.random_seed

    s = pickle.dumps(ic_)

    ic2 = pickle.loads(s)
    assert repr(ic_) == repr(ic2)


def test_fname(default_user_params):
    ic1 = InitialConditions(user_params=default_user_params)
    ic2 = InitialConditions(user_params=default_user_params)

    # we didn't give them seeds, so can't access the filename attribute
    # (it is undefined until a seed is set)
    with pytest.raises(AttributeError):
        assert ic1.filename != ic2.filename  # random seeds are different

    # *but* should be able to get a skeleton filename:
    assert ic1._fname_skeleton == ic2._fname_skeleton

    ic1.random_seed  # sets the random seed
    ic2.random_seed

    assert ic1.filename != ic2.filename  # random seeds should now be different
    assert ic1._fname_skeleton == ic2._fname_skeleton


def test_match_seed(tmpdirec, default_user_params):
    ic2 = InitialConditions(random_seed=1, user_params=default_user_params)

    # This fails because we've set the seed and it's different to the existing one.
    with pytest.raises(IOError):
        ic2.read(direc=tmpdirec)


def test_bad_class_definition(default_user_params):
    class CustomInitialConditions(InitialConditions):
        _name = "InitialConditions"

        """
        A class containing all initial conditions boxes.
        """

        def _get_box_structures(self):
            out = super()._get_box_structures()
            out["unknown_key"] = (1, 1, 1)
            return out

    with pytest.raises(TypeError):
        CustomInitialConditions(init=True, user_params=default_user_params)


def test_bad_write(init):
    # no random seed yet so shouldn't be able to write.
    with pytest.raises(IOError):
        init.write()


def test_global_params_keys():
    assert "HII_FILTER" in global_params.keys()


def test_reading_purged(ic: InitialConditions):
    lowres_density = ic.lowres_density

    # Remove it from memory
    ic.purge()

    assert "lowres_density" not in ic.__dict__
    assert ic._array_state["lowres_density"].on_disk
    assert not ic._array_state["lowres_density"].computed_in_mem

    # But we can still get it.
    lowres_density_2 = ic.lowres_density

    assert ic._array_state["lowres_density"].on_disk
    assert ic._array_state["lowres_density"].computed_in_mem

    assert np.allclose(lowres_density_2, lowres_density)

    ic.load_all()



import pytest

from py21cmfast._utils import PHOTONCONSERROR, ParameterError
from py21cmfast.c_21cmfast import lib
from py21cmfast.wrapper import _call_c_simple


@pytest.mark.parametrize("subfunc", [True, False])
def test_basic(subfunc):
    status = lib.SomethingThatCatches(subfunc)
    assert status == PHOTONCONSERROR


@pytest.mark.parametrize("subfunc", [True, False])
def test_simple(subfunc):
    with pytest.raises(ParameterError):
        _call_c_simple(lib.FunctionThatCatches, subfunc, False)


def test_pass():
    answer = _call_c_simple(lib.FunctionThatCatches, True, True)
    assert answer == 5.0



import pytest

import logging
import os

from py21cmfast import UserParams, config, global_params, run_lightcone, wrapper
from py21cmfast.cache_tools import clear_cache


def pytest_addoption(parser):
    parser.addoption("--log-level-21", action="store", default="WARNING")


@pytest.fixture(scope="session")
def tmpdirec(tmp_path_factory):
    """Pytest fixture instantiating a new session-scope "data" folder.

    Parameters
    ----------
    tmpdir_factory :
        Pytest fixture for creating temporary directories.
    """
    return tmp_path_factory.mktemp("data")


def printdir(direc):
    try:
        width = os.get_terminal_size().columns
    except OSError:
        width = 100

    print()
    print(f" Files In {direc} ".center(width, "="))
    for pth in direc.iterdir():
        print(f"\t {pth.name:<20}:\t\t {pth.stat().st_size / 1024**2:.3f} KB")
    print("=" * width)


@pytest.fixture(scope="module")
def module_direc(tmp_path_factory):

    original = config["direc"]
    direc = tmp_path_factory.mktemp("modtmp")

    config["direc"] = str(direc)

    yield direc

    printdir(direc)

    # Clear all cached items created.
    clear_cache(direc=str(direc))

    # Set direc back to original.
    config["direc"] = original


@pytest.fixture(scope="function")
def test_direc(tmp_path_factory):

    original = config["direc"]
    direc = tmp_path_factory.mktemp("testtmp")

    config["direc"] = str(direc)

    yield direc

    printdir(direc)
    # Clear all cached items created.
    clear_cache(direc=str(direc))

    # Set direc back to original.
    config["direc"] = original


@pytest.fixture(autouse=True, scope="session")
def setup_and_teardown_package(tmpdirec, request):
    # Set nice global defaults for testing purposes, to make runs faster
    # (can always be over-ridden per-test).
    original_zprime = global_params.ZPRIME_STEP_FACTOR

    # Set default global parameters for all tests
    global_params.ZPRIME_STEP_FACTOR = 1.2

    # Set default config parameters for all tests.
    config["direc"] = str(tmpdirec)
    config["regenerate"] = True
    config["write"] = False

    log_level = request.config.getoption("--log-level-21") or logging.INFO
    logging.getLogger("py21cmfast").setLevel(log_level)

    yield

    printdir(tmpdirec)

    clear_cache(direc=str(tmpdirec))
    global_params.ZPRIME_STEP_FACTOR = original_zprime


# ======================================================================================
# Create a default set of boxes that can be used throughout.
# ======================================================================================


@pytest.fixture(scope="session")
def default_user_params():
    return UserParams(HII_DIM=35, DIM=70, BOX_LEN=50)


@pytest.fixture(scope="session")
def ic(default_user_params, tmpdirec):
    return wrapper.initial_conditions(
        user_params=default_user_params, write=True, direc=tmpdirec, random_seed=12
    )


@pytest.fixture(scope="session")
def redshift():
    """A default redshift to evaluate at. Not too high, not too low."""
    return 15


@pytest.fixture(scope="session")
def max_redshift():
    """A default redshift to evaluate at. Not too high, not too low."""
    return 25


@pytest.fixture(scope="session")
def low_redshift():
    """A default redshift to evaluate at. Not too high, not too low."""
    return 8


@pytest.fixture(scope="session")
def perturb_field(ic, redshift):
    """A default perturb_field"""
    return wrapper.perturb_field(redshift=redshift, init_boxes=ic, write=True)


@pytest.fixture(scope="session")
def lc(perturb_field, max_redshift):
    return run_lightcone(perturb=perturb_field, max_redshift=max_redshift)



"""A set of tools for reading/writing/querying the in-built cache."""
import glob
import h5py
import logging
import os
import re
from os import path

from . import outputs, wrapper
from ._cfg import config
from .wrapper import global_params

logger = logging.getLogger("21cmFAST")


def readbox(
    *,
    direc=None,
    fname=None,
    hsh=None,
    kind=None,
    seed=None,
    redshift=None,
    load_data=True,
):
    """
    Read in a data set and return an appropriate object for it.

    Parameters
    ----------
    direc : str, optional
        The directory in which to search for the boxes. By default, this is the
        centrally-managed directory, given by the ``config.yml`` in ``~/.21cmfast/``.
    fname: str, optional
        The filename (without directory) of the data set. If given, this will be
        preferentially used, and must exist.
    hsh: str, optional
        The md5 hsh of the object desired to be read. Required if `fname` not given.
    kind: str, optional
        The kind of dataset, eg. "InitialConditions". Will be the name of a class
        defined in :mod:`~wrapper`. Required if `fname` not given.
    seed: str or int, optional
        The random seed of the data set to be read. If not given, and filename not
        given, then a box will be read if it matches the kind and hsh, with an
        arbitrary seed.
    load_data: bool, optional
        Whether to read in the data in the data set. Otherwise, only its defining
        parameters are read.

    Returns
    -------
    dataset :
        An output object, whose type depends on the kind of data set being read.

    Raises
    ------
    IOError :
        If no files exist of the given kind and hsh.
    ValueError :
        If either ``fname`` is not supplied, or both ``kind`` and ``hsh`` are not supplied.
    """
    direc = path.expanduser(direc or config["direc"])

    if not (fname or (hsh and kind)):
        raise ValueError("Either fname must be supplied, or kind and hsh")

    zstr = f"z{redshift:.4f}_" if redshift is not None else ""
    if not fname:
        if not seed:
            fname = kind + "_" + zstr + hsh + "_r*.h5"
            files = glob.glob(path.join(direc, fname))
            if files:
                fname = files[0]
            else:
                raise OSError("No files exist with that kind and hsh.")
        else:
            fname = kind + "_" + zstr + hsh + "_r" + str(seed) + ".h5"

    kind = _parse_fname(fname)["kind"]
    cls = getattr(outputs, kind)

    if hasattr(cls, "from_file"):
        inst = cls.from_file(fname, direc=direc, load_data=load_data)
    else:
        inst = cls.read(fname, direc=direc)

    return inst


def _parse_fname(fname):
    patterns = (
        r"(?P<kind>\w+)_(?P<hash>\w{32})_r(?P<seed>\d+).h5$",
        r"(?P<kind>\w+)_z(?P<redshift>\d+.\d{1,4})_(?P<hash>\w{32})_r(?P<seed>\d+).h5$",
    )
    for pattern in patterns:
        match = re.match(pattern, os.path.basename(fname))
        if match:
            break

    if not match:
        raise ValueError(
            "filename {} does not have correct format for a cached output.".format(
                fname
            )
        )

    return match.groupdict()


def list_datasets(*, direc=None, kind=None, hsh=None, seed=None, redshift=None):
    """Yield all datasets which match a given set of filters.

    Can be used to determine parameters of all cached datasets, in conjunction with :func:`readbox`.

    Parameters
    ----------
    direc : str, optional
        The directory in which to search for the boxes. By default, this is the centrally-managed
        directory, given by the ``config.yml`` in ``.21cmfast``.
    kind: str, optional
        Filter by this kind (one of {"InitialConditions", "PerturbedField", "IonizedBox",
        "TsBox", "BrightnessTemp"}
    hsh: str, optional
        Filter by this hsh.
    seed: str, optional
        Filter by this seed.

    Yields
    ------
    fname: str
        The filename of the dataset (without directory).
    parts: tuple of strings
        The (kind, hsh, seed) of the data set.
    """
    direc = path.expanduser(direc or config["direc"])

    fname = "{}{}_{}_r{}.h5".format(
        kind or r"(?P<kind>[a-zA-Z]+)",
        f"_z{redshift:.4f}" if redshift is not None else "(.*)",
        hsh or r"(?P<hash>\w{32})",
        seed or r"(?P<seed>\d+)",
    )

    for fl in os.listdir(direc):
        if re.match(fname, fl):
            yield fl


def query_cache(
    *, direc=None, kind=None, hsh=None, seed=None, redshift=None, show=True
):
    """Get or print datasets in the cache.

    Walks through the cache, with given filters, and return all un-initialised dataset
    objects, optionally printing their representation to screen.
    Useful for querying which kinds of datasets are available within the cache, and
    choosing one to read and use.

    Parameters
    ----------
    direc : str, optional
        The directory in which to search for the boxes. By default, this is the
        centrally-managed directory, given by the ``config.yml`` in ``~/.21cmfast``.
    kind: str, optional
        Filter by this kind. Must be one of "InitialConditions", "PerturbedField",
        "IonizedBox", "TsBox" or "BrightnessTemp".
    hsh: str, optional
        Filter by this hsh.
    seed: str, optional
        Filter by this seed.
    show: bool, optional
        Whether to print out a repr of each object that exists.

    Yields
    ------
    obj:
       Output objects, un-initialized.
    """
    for file in list_datasets(
        direc=direc, kind=kind, hsh=hsh, seed=seed, redshift=redshift
    ):
        cls = readbox(direc=direc, fname=file, load_data=False)
        if show:
            print(file + ": " + str(cls))  # noqa: T
        yield file, cls


def clear_cache(**kwargs):
    """Delete datasets in the cache.

    Walks through the cache, with given filters, and deletes all un-initialised dataset
    objects, optionally printing their representation to screen.

    Parameters
    ----------
    kwargs :
        All options passed through to :func:`query_cache`.
    """
    direc = kwargs.get("direc", path.expanduser(config["direc"]))
    number = 0
    for fname, cls in query_cache(show=False, **kwargs):
        if kwargs.get("show", True):
            logger.info(f"Removing {fname}")
        os.remove(path.join(direc, fname))
        number += 1

    logger.info(f"Removed {number} files from cache.")



import pytest

import yaml
from os import path

import py21cmfast as p21
from py21cmfast._cfg import Config


@pytest.fixture(scope="module")
def cfgdir(tmp_path_factory):
    return tmp_path_factory.mktemp("config_test_dir")


def test_config_context(cfgdir, default_user_params):
    with p21.config.use(direc=cfgdir, write=True):
        init = p21.initial_conditions(user_params=default_user_params)

    assert (cfgdir / init.filename).exists()
    assert "config_test_dir" not in p21.config["direc"]


def test_config_write(cfgdir):
    with p21.config.use(direc=str(cfgdir)):
        p21.config.write(cfgdir / "config.yml")

    with open(cfgdir / "config.yml") as fl:
        new_config = yaml.load(fl, Loader=yaml.FullLoader)

    # Test adding new kind of string alias
    new_config["boxdir"] = new_config["direc"]
    del new_config["direc"]

    with open(cfgdir / "config.yml", "w") as fl:
        yaml.dump(new_config, fl)

    with pytest.warns(UserWarning):
        new_config = Config.load(cfgdir / "config.yml")

    assert "boxdir" not in new_config
    assert "direc" in new_config

    with open(cfgdir / "config.yml") as fl:
        new_config = yaml.load(fl, Loader=yaml.FullLoader)

    assert "boxdir" not in new_config
    assert "direc" in new_config



"""Simple script to simulate an MCMC-like routine to check for memory leaks."""
import numpy as np
import os
import psutil
import tracemalloc

from py21cmfast import initial_conditions, perturb_field, run_lightcone

tracemalloc.start()
snapshot = tracemalloc.take_snapshot()
PROCESS = psutil.Process(os.getpid())
oldmem = 0


def trace_print():
    """Print a trace of memory leaks."""
    global snapshot
    global oldmem

    snapshot2 = tracemalloc.take_snapshot()
    snapshot2 = snapshot2.filter_traces(
        (
            tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
            tracemalloc.Filter(False, "<unknown>"),
            tracemalloc.Filter(False, tracemalloc.__file__),
        )
    )

    if snapshot is not None:
        thismem = PROCESS.memory_info().rss / 1024**2
        diff = thismem - oldmem
        print(
            "===================== Begin Trace (TOTAL MEM={:1.4e} MB... [{:+1.4e} MB]):".format(
                thismem, diff
            )
        )
        top_stats = snapshot2.compare_to(snapshot, "lineno", cumulative=True)
        for stat in top_stats[:4]:
            print(stat)
        print("End Trace ===========================================")
        oldmem = thismem

    snapshot = snapshot2


trace_print()

run_lightcone(
    redshift=15,
    user_params={
        "USE_INTERPOLATION_TABLES": True,
        "N_THREADS": 1,
        "DIM": 100,
        "HII_DIM": 25,
        "PERTURB_ON_HIGH_RES": True,
        "USE_FFTW_WISDOM": False,
    },
    flag_options={
        "USE_MASS_DEPENDENT_ZETA": True,
        "INHOMO_RECO": True,
        "USE_TS_FLUCT": True,
    },
    direc="_cache_%s" % (os.path.basename(__file__)[:-3]),
    random_seed=1993,
)



"""Quick test of finding halos."""

from py21cmfast import (
    AstroParams,
    CosmoParams,
    FlagOptions,
    UserParams,
    determine_halo_list,
    initial_conditions,
    perturb_field,
)
from py21cmfast._utils import StructInstanceWrapper
from py21cmfast.c_21cmfast import ffi, lib

global_params = StructInstanceWrapper(lib.global_params, ffi)

user_params = UserParams(
    DIM=150,
    HII_DIM=50,
    BOX_LEN=150.0,
    USE_FFTW_WISDOM=False,
    HMF=1,
    N_THREADS=1,
    NO_RNG=True,
    PERTURB_ON_HIGH_RES=True,
)

flag_options = FlagOptions(
    USE_MASS_DEPENDENT_ZETA=True,
    USE_TS_FLUCT=False,
    INHOMO_RECO=False,
    SUBCELL_RSD=False,
    M_MIN_in_Mass=False,
    PHOTON_CONS=False,
)

if __name__ == "__main__":
    random_seed = 42

    cosmo_params = CosmoParams(
        OMb=0.0486, OMm=0.3075, POWER_INDEX=0.97, SIGMA_8=0.82, hlittle=0.6774
    )

    astro_params = AstroParams(
        ALPHA_ESC=-0.5,
        ALPHA_STAR=0.5,
        F_ESC10=-1.30102999566,
        F_STAR10=-1.0,
        L_X=40.5,
        M_TURN=8.7,
        NU_X_THRESH=500.0,
        X_RAY_SPEC_INDEX=1.0,
        t_STAR=0.5,
        R_BUBBLE_MAX=15.0,
    )

    init_box = initial_conditions(
        user_params=user_params,
        cosmo_params=cosmo_params,
        random_seed=random_seed,
        regenerate=True,
        write=False,
    )

    redshift = 9.0

    pt_box = perturb_field(
        redshift=redshift,
        init_boxes=init_box,
        user_params=user_params,
        cosmo_params=cosmo_params,
    )

    halos = determine_halo_list(
        redshift=redshift,
        init_boxes=init_box,
        user_params=user_params,
        cosmo_params=cosmo_params,
        astro_params=astro_params,
        flag_options=flag_options,
        regenerate=True,
        write=False,
        OPTIMIZE=False,
    )

    print(halos.halo_masses)



"""The py21cmfast package."""
try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:
    from importlib_metadata import PackageNotFoundError, version

try:
    from ._version import version as __version__
except ModuleNotFoundError:  # pragma: no cover
    try:
        __version__ = version("21cmFAST")
    except PackageNotFoundError:
        # package is not installed
        __version__ = "unknown"

# This just ensures that the default directory for boxes is created.
from os import mkdir as _mkdir
from os import path

from . import cache_tools, inputs, outputs, plotting, wrapper
from ._cfg import config
from ._logging import configure_logging
from .cache_tools import query_cache
from .outputs import (
    Coeval,
    InitialConditions,
    IonizedBox,
    LightCone,
    PerturbedField,
    TsBox,
)
from .wrapper import (
    AstroParams,
    BrightnessTemp,
    CosmoParams,
    FlagOptions,
    HaloField,
    PerturbHaloField,
    UserParams,
    brightness_temperature,
    compute_luminosity_function,
    compute_tau,
    construct_fftw_wisdoms,
    determine_halo_list,
    get_all_fieldnames,
    global_params,
    initial_conditions,
    ionize_box,
    perturb_field,
    perturb_halo_list,
    run_coeval,
    run_lightcone,
    spin_temperature,
)

configure_logging()

try:
    _mkdir(path.expanduser(config["direc"]))
except FileExistsError:
    pass



import pytest

import h5py
import numpy as np

from py21cmfast import (
    BrightnessTemp,
    Coeval,
    LightCone,
    TsBox,
    global_params,
    run_coeval,
    run_lightcone,
)


@pytest.fixture(scope="module")
def coeval(ic):
    return run_coeval(
        redshift=25.0, init_box=ic, write=True, flag_options={"USE_TS_FLUCT": True}
    )


@pytest.fixture(scope="module")
def lightcone(ic):
    return run_lightcone(
        max_redshift=35,
        redshift=25.0,
        init_box=ic,
        write=True,
        flag_options={"USE_TS_FLUCT": True},
    )


def test_lightcone_roundtrip(test_direc, lc):
    fname = lc.save(direc=test_direc)
    lc2 = LightCone.read(fname)

    assert lc == lc2
    assert lc.get_unique_filename() == lc2.get_unique_filename()
    assert np.all(np.isclose(lc.brightness_temp, lc2.brightness_temp))


def test_lightcone_io_abspath(lc, test_direc):
    lc.save(test_direc / "abs_path_lightcone.h5")
    assert (test_direc / "abs_path_lightcone.h5").exists()


def test_coeval_roundtrip(test_direc, coeval):
    fname = coeval.save(direc=test_direc)
    coeval2 = Coeval.read(fname)

    assert coeval == coeval2
    assert coeval.get_unique_filename() == coeval2.get_unique_filename()
    assert np.all(
        np.isclose(
            coeval.brightness_temp_struct.brightness_temp,
            coeval2.brightness_temp_struct.brightness_temp,
        )
    )


def test_coeval_cache(coeval):
    assert coeval.cache_files is not None
    out = coeval.get_cached_data(kind="brightness_temp", redshift=25.1, load_data=True)

    assert isinstance(out, BrightnessTemp)
    assert np.all(out.brightness_temp == coeval.brightness_temp)

    out = coeval.get_cached_data(
        kind="spin_temp", redshift=global_params.Z_HEAT_MAX * 1.01, load_data=True
    )

    assert isinstance(out, TsBox)
    assert not np.all(out.Ts_box == coeval.Ts_box)

    with pytest.raises(ValueError):
        coeval.get_cached_data(kind="bad", redshift=100.0)


def test_gather(coeval, test_direc):
    fname = coeval.gather(
        fname="tmpfile_test_gather.h5",
        kinds=("perturb_field", "init"),
        direc=str(test_direc),
    )

    with h5py.File(fname, "r") as fl:
        assert "cache" in fl
        assert "perturb_field" in fl["cache"]
        assert "z25.00" in fl["cache"]["perturb_field"]
        assert "density" in fl["cache"]["perturb_field"]["z25.00"]
        assert (
            fl["cache"]["perturb_field"]["z25.00"]["density"].shape
            == (coeval.user_params.HII_DIM,) * 3
        )

        assert "z0.00" in fl["cache"]["init"]


def test_lightcone_cache(lightcone):
    assert lightcone.cache_files is not None
    out = lightcone.get_cached_data(
        kind="brightness_temp", redshift=25.1, load_data=True
    )

    assert isinstance(out, BrightnessTemp)

    out = lightcone.get_cached_data(
        kind="brightness_temp", redshift=global_params.Z_HEAT_MAX * 1.01, load_data=True
    )

    assert isinstance(out, BrightnessTemp)
    assert out.redshift != lightcone.redshift

    with pytest.raises(ValueError):
        lightcone.get_cached_data(kind="bad", redshift=100.0)

    print(lightcone.cache_files)
    lightcone.gather(fname="tmp_lightcone_gather.h5", clean=["brightness_temp"])

    with pytest.raises(IOError):
        lightcone.get_cached_data(kind="brightness_temp", redshift=25.1)



"""
Various tests of the initial_conditions() function and InitialConditions class.
"""
import pytest

import numpy as np
from multiprocessing import cpu_count

from py21cmfast import wrapper


def test_box_shape(ic):
    """Test basic properties of the InitialConditions struct"""
    shape = (35, 35, 35)
    hires_shape = tuple(2 * s for s in shape)
    assert ic.lowres_density.shape == shape
    assert ic.lowres_vx.shape == shape
    assert ic.lowres_vy.shape == shape
    assert ic.lowres_vz.shape == shape
    assert ic.lowres_vx_2LPT.shape == shape
    assert ic.lowres_vy_2LPT.shape == shape
    assert ic.lowres_vz_2LPT.shape == shape
    assert ic.hires_density.shape == hires_shape

    assert ic.hires_vx.shape == hires_shape
    assert ic.hires_vy.shape == hires_shape
    assert ic.hires_vz.shape == hires_shape
    assert ic.hires_vx_2LPT.shape == hires_shape
    assert ic.hires_vy_2LPT.shape == hires_shape
    assert ic.hires_vz_2LPT.shape == hires_shape

    assert not hasattr(ic, "lowres_vcb")

    assert ic.cosmo_params == wrapper.CosmoParams()


def test_modified_cosmo(ic):
    """Test using a modified cosmology"""
    cosmo = wrapper.CosmoParams(SIGMA_8=0.9)
    ic2 = wrapper.initial_conditions(
        cosmo_params=cosmo,
        user_params=ic.user_params,
    )

    assert ic2.cosmo_params != ic.cosmo_params
    assert ic2.cosmo_params == cosmo
    assert ic2.cosmo_params.SIGMA_8 == cosmo.SIGMA_8


def test_transfer_function(ic, default_user_params):
    """Test using a modified transfer function"""
    user_params = default_user_params.clone(POWER_SPECTRUM=5)
    ic2 = wrapper.initial_conditions(
        random_seed=ic.random_seed,
        user_params=user_params,
    )

    rmsnew = np.sqrt(np.mean(ic2.hires_density**2))
    rmsdelta = np.sqrt(np.mean((ic2.hires_density - ic.hires_density) ** 2))
    assert rmsdelta < rmsnew
    assert rmsnew > 0.0
    assert not np.allclose(ic2.hires_density, ic.hires_density)


def test_relvels():
    """Test for relative velocity initial conditions"""
    ic = wrapper.initial_conditions(
        random_seed=1,
        user_params=wrapper.UserParams(
            HII_DIM=100,
            DIM=300,
            BOX_LEN=300,
            POWER_SPECTRUM=5,
            USE_RELATIVE_VELOCITIES=True,
            N_THREADS=cpu_count(),  # To make this one a bit faster.
        ),
    )

    vcbrms_lowres = np.sqrt(np.mean(ic.lowres_vcb**2))
    vcbavg_lowres = np.mean(ic.lowres_vcb)

    # we test the lowres box
    # rms should be about 30 km/s for LCDM, so we check it is finite and not far off
    # the average should be 0.92*vrms, since it follows a maxwell boltzmann
    assert vcbrms_lowres > 20.0
    assert vcbrms_lowres < 40.0
    assert vcbavg_lowres < 0.97 * vcbrms_lowres
    assert vcbavg_lowres > 0.88 * vcbrms_lowres



"""Utilities that help with wrapping various C structures."""
import glob
import h5py
import logging
import numpy as np
import warnings
from abc import ABCMeta, abstractmethod
from bidict import bidict
from cffi import FFI
from enum import IntEnum
from hashlib import md5
from os import makedirs, path
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

from . import __version__
from ._cfg import config
from .c_21cmfast import lib

_ffi = FFI()

logger = logging.getLogger(__name__)


class ArrayStateError(ValueError):
    """Errors arising from incorrectly modifying array state."""

    pass


class ArrayState:
    """Define the memory state of a struct array."""

    def __init__(
        self, initialized=False, c_memory=False, computed_in_mem=False, on_disk=False
    ):
        self._initialized = initialized
        self._c_memory = c_memory
        self._computed_in_mem = computed_in_mem
        self._on_disk = on_disk

    @property
    def initialized(self):
        """Whether the array is initialized (i.e. allocated memory)."""
        return self._initialized

    @initialized.setter
    def initialized(self, val):
        if not val:
            # if its not initialized, can't be computed in memory
            self.computed_in_mem = False
        self._initialized = bool(val)

    @property
    def c_memory(self):
        """Whether the array's memory (if any) is controlled by C."""
        return self._c_memory

    @c_memory.setter
    def c_memory(self, val):
        self._c_memory = bool(val)

    @property
    def computed_in_mem(self):
        """Whether the array is computed and stored in memory."""
        return self._computed_in_mem

    @computed_in_mem.setter
    def computed_in_mem(self, val):
        if val:
            # any time we pull something into memory, it must be initialized.
            self.initialized = True
        self._computed_in_mem = bool(val)

    @property
    def on_disk(self):
        """Whether the array is computed and store on disk."""
        return self._on_disk

    @on_disk.setter
    def on_disk(self, val):
        self._on_disk = bool(val)

    @property
    def computed(self):
        """Whether the array is computed anywhere."""
        return self.computed_in_mem or self.on_disk

    @property
    def c_has_active_memory(self):
        """Whether C currently has initialized memory for this array."""
        return self.c_memory and self.initialized


class ParameterError(RuntimeError):
    """An exception representing a bad choice of parameters."""

    default_message = "21cmFAST does not support this combination of parameters."

    def __init__(self, msg=None):
        super().__init__(msg or self.default_message)


class FatalCError(Exception):
    """An exception representing something going wrong in C."""

    default_message = "21cmFAST is exiting."

    def __init__(self, msg=None):
        super().__init__(msg or self.default_message)


class FileIOError(FatalCError):
    """An exception when an error occurs with file I/O."""

    default_message = "Expected file could not be found! (check the LOG for more info)"


class GSLError(ParameterError):
    """An exception when a GSL routine encounters an error."""

    default_message = "A GSL routine has errored! (check the LOG for more info)"


class ArgumentValueError(FatalCError):
    """An exception when a function takes an unexpected input."""

    default_message = "An incorrect argument has been defined or passed! (check the LOG for more info)"


class PhotonConsError(ParameterError):
    """An exception when the photon non-conservation correction routine errors."""

    default_message = "An error has occured with the Photon non-conservation correction! (check the LOG for more info)"


class TableGenerationError(ParameterError):
    """An exception when an issue arises populating one of the interpolation tables."""

    default_message = """An error has occured when generating an interpolation table!
                This has likely occured due to the choice of input AstroParams (check the LOG for more info)"""


class TableEvaluationError(ParameterError):
    """An exception when an issue arises populating one of the interpolation tables."""

    default_message = """An error has occured when evaluating an interpolation table!
                This can sometimes occur due to small boxes (either small DIM/HII_DIM or BOX_LEN) (check the LOG for more info)"""


class InfinityorNaNError(ParameterError):
    """An exception when an infinity or NaN is encountered in a calculated quantity."""

    default_message = """Something has returned an infinity or a NaN! This could be due to an issue with an
                input parameter choice (check the LOG for more info)"""


class MassDepZetaError(ParameterError):
    """An exception when determining the bisection for stellar mass/escape fraction."""

    default_message = """There is an issue with the choice of parameters under MASS_DEPENDENT_ZETA. Could be an issue with
                any of the chosen F_STAR10, ALPHA_STAR, F_ESC10 or ALPHA_ESC."""


class MemoryAllocError(FatalCError):
    """An exception when unable to allocated memory."""

    default_message = """An error has occured while attempting to allocate memory! (check the LOG for more info)"""


SUCCESS = 0
IOERROR = 1
GSLERROR = 2
VALUEERROR = 3
PHOTONCONSERROR = 4
TABLEGENERATIONERROR = 5
TABLEEVALUATIONERROR = 6
INFINITYORNANERROR = 7
MASSDEPZETAERROR = 8
MEMORYALLOCERROR = 9


def _process_exitcode(exitcode, fnc, args):
    """Determine what happens for different values of the (integer) exit code from a C function."""
    if exitcode != SUCCESS:
        logger.error(f"In function: {fnc.__name__}.  Arguments: {args}")

        if exitcode:
            try:
                raise {
                    IOERROR: FileIOError,
                    GSLERROR: GSLError,
                    VALUEERROR: ArgumentValueError,
                    PHOTONCONSERROR: PhotonConsError,
                    TABLEGENERATIONERROR: TableGenerationError,
                    TABLEEVALUATIONERROR: TableEvaluationError,
                    MASSDEPZETAERROR: MassDepZetaError,
                    MEMORYALLOCERROR: MemoryAllocError,
                }[exitcode]
            except KeyError:  # pragma: no cover
                raise FatalCError(
                    "Unknown error in C. Please report this error!"
                )  # Unknown C code


ctype2dtype = {}

# Integer types
for prefix in ("int", "uint"):
    for log_bytes in range(4):
        ctype = "%s%d_t" % (prefix, 8 * (2**log_bytes))
        dtype = "%s%d" % (prefix[0], 2**log_bytes)
        ctype2dtype[ctype] = np.dtype(dtype)

# Floating point types
ctype2dtype["float"] = np.dtype("f4")
ctype2dtype["double"] = np.dtype("f8")
ctype2dtype["int"] = np.dtype("i4")


def asarray(ptr, shape):
    """Get the canonical C type of the elements of ptr as a string."""
    ctype = _ffi.getctype(_ffi.typeof(ptr).item).split("*")[0].strip()

    if ctype not in ctype2dtype:
        raise RuntimeError(
            f"Cannot create an array for element type: {ctype}. Can do {list(ctype2dtype.values())}."
        )

    array = np.frombuffer(
        _ffi.buffer(ptr, _ffi.sizeof(ctype) * np.prod(shape)), ctype2dtype[ctype]
    )
    array.shape = shape
    return array


class StructWrapper:
    """
    A base-class python wrapper for C structures (not instances of them).

    Provides simple methods for creating new instances and accessing field names and values.

    To implement wrappers of specific structures, make a subclass with the same name as the
    appropriate C struct (which must be defined in the C code that has been compiled to the ``ffi``
    object), *or* use an arbitrary name, but set the ``_name`` attribute to the C struct name.
    """

    _name = None
    _ffi = None

    def __init__(self):

        # Set the name of this struct in the C code
        self._name = self._get_name()

    @classmethod
    def _get_name(cls):
        return cls._name or cls.__name__

    @property
    def _cstruct(self):
        """
        The actual structure which needs to be passed around to C functions.

        .. note:: This is best accessed by calling the instance (see __call__).

        The reason it is defined as this (manual) cached property is so that it can be created
        dynamically, but not lost. It must not be lost, or else C functions which use it will lose
        access to its memory. But it also must be created dynamically so that it can be recreated
        after pickling (pickle can't handle CData).
        """
        try:
            return self.__cstruct
        except AttributeError:
            self.__cstruct = self._new()
            return self.__cstruct

    def _new(self):
        """Return a new empty C structure corresponding to this class."""
        return self._ffi.new("struct " + self._name + "*")

    @classmethod
    def get_fields(cls, cstruct=None) -> Dict[str, Any]:
        """Obtain the C-side fields of this struct."""
        if cstruct is None:
            cstruct = cls._ffi.new("struct " + cls._get_name() + "*")
        return dict(cls._ffi.typeof(cstruct[0]).fields)

    @classmethod
    def get_fieldnames(cls, cstruct=None) -> List[str]:
        """Obtain the C-side field names of this struct."""
        fields = cls.get_fields(cstruct)
        return [f for f, t in fields]

    @classmethod
    def get_pointer_fields(cls, cstruct=None) -> List[str]:
        """Obtain all pointer fields of the struct (typically simulation boxes)."""
        return [f for f, t in cls.get_fields(cstruct) if t.type.kind == "pointer"]

    @property
    def fields(self) -> Dict[str, Any]:
        """List of fields of the underlying C struct (a list of tuples of "name, type")."""
        return self.get_fields(self._cstruct)

    @property
    def fieldnames(self) -> List[str]:
        """List names of fields of the underlying C struct."""
        return [f for f, t in self.fields.items()]

    @property
    def pointer_fields(self) -> List[str]:
        """List of names of fields which have pointer type in the C struct."""
        return [f for f, t in self.fields.items() if t.type.kind == "pointer"]

    @property
    def primitive_fields(self) -> List[str]:
        """List of names of fields which have primitive type in the C struct."""
        return [f for f, t in self.fields.items() if t.type.kind == "primitive"]

    def __getstate__(self):
        """Return the current state of the class without pointers."""
        return {
            k: v
            for k, v in self.__dict__.items()
            if k not in ["_strings", "_StructWrapper__cstruct"]
        }

    def refresh_cstruct(self):
        """Delete the underlying C object, forcing it to be rebuilt."""
        try:
            del self.__cstruct
        except AttributeError:
            pass

    def __call__(self):
        """Return an instance of the C struct."""
        pass


class StructWithDefaults(StructWrapper):
    """
    A convenient interface to create a C structure with defaults specified.

    It is provided for the purpose of *creating* C structures in Python to be passed to C functions,
    where sensible defaults are available. Structures which are created within C and passed back do
    not need to be wrapped.

    This provides a *fully initialised* structure, and will fail if not all fields are specified
    with defaults.

    .. note:: The actual C structure is gotten by calling an instance. This is auto-generated when
              called, based on the parameters in the class.

    .. warning:: This class will *not* deal well with parameters of the struct which are pointers.
                 All parameters should be primitive types, except for strings, which are dealt with
                 specially.

    Parameters
    ----------
    ffi : cffi object
        The ffi object from any cffi-wrapped library.
    """

    _defaults_ = {}

    def __init__(self, *args, **kwargs):

        super().__init__()

        if args:
            if len(args) > 1:
                raise TypeError(
                    "%s takes up to one position argument, %s were given"
                    % (self.__class__.__name__, len(args))
                )
            elif args[0] is None:
                pass
            elif isinstance(args[0], self.__class__):
                kwargs.update(args[0].self)
            elif isinstance(args[0], dict):
                kwargs.update(args[0])
            else:
                raise TypeError(
                    f"optional positional argument for {self.__class__.__name__} must be"
                    f" None, dict, or an instance of itself"
                )

        for k, v in self._defaults_.items():

            # Prefer arguments given to the constructor.
            _v = kwargs.pop(k, None)

            if _v is not None:
                v = _v

            try:
                setattr(self, k, v)
            except AttributeError:
                # The attribute has been defined as a property, save it as a hidden variable
                setattr(self, "_" + k, v)

        if kwargs:
            logger.warning(
                "The following parameters to {thisclass} are not supported: {lst}".format(
                    thisclass=self.__class__.__name__, lst=list(kwargs.keys())
                )
            )

    def convert(self, key, val):
        """Make any conversions of values before saving to the instance."""
        return val

    def update(self, **kwargs):
        """
        Update the parameters of an existing class structure.

        This should always be used instead of attempting to *assign* values to instance attributes.
        It consistently re-generates the underlying C memory space and sets some book-keeping
        variables.

        Parameters
        ----------
        kwargs:
            Any argument that may be passed to the class constructor.
        """
        # Start a fresh cstruct.
        if kwargs:
            self.refresh_cstruct()

        for k in self._defaults_:
            # Prefer arguments given to the constructor.
            if k in kwargs:
                v = kwargs.pop(k)

                try:
                    setattr(self, k, v)
                except AttributeError:
                    # The attribute has been defined as a property, save it as a hidden variable
                    setattr(self, "_" + k, v)

        # Also ensure that parameters that are part of the class, but not the defaults, are set
        # this will fail if these parameters cannot be set for some reason, hence doing it
        # last.
        for k in list(kwargs.keys()):
            if hasattr(self, k):
                setattr(self, k, kwargs.pop(k))

        if kwargs:
            warnings.warn(
                "The following arguments to be updated are not compatible with this class: %s"
                % kwargs
            )

    def clone(self, **kwargs):
        """Make a fresh copy of the instance with arbitrary parameters updated."""
        new = self.__class__(self.self)
        new.update(**kwargs)
        return new

    def __call__(self):
        """Return a filled C Structure corresponding to this instance."""
        for key, val in self.pystruct.items():

            # Find the value of this key in the current class
            if isinstance(val, str):
                # If it is a string, need to convert it to C string ourselves.
                val = self.ffi.new("char[]", getattr(self, key).encode())

            try:
                setattr(self._cstruct, key, val)
            except TypeError:
                logger.info(f"For key {key}, value {val}:")
                raise

        return self._cstruct

    @property
    def pystruct(self):
        """A pure-python dictionary representation of the corresponding C structure."""
        return {fld: self.convert(fld, getattr(self, fld)) for fld in self.fieldnames}

    @property
    def defining_dict(self):
        """
        Pure python dictionary representation of this class, as it would appear in C.

        .. note:: This is not the same as :attr:`pystruct`, as it omits all variables that don't
                  need to be passed to the constructor, but appear in the C struct (some can be
                  calculated dynamically based on the inputs). It is also not the same as
                  :attr:`self`, as it includes the 'converted' values for each variable, which are
                  those actually passed to the C code.
        """
        return {k: self.convert(k, getattr(self, k)) for k in self._defaults_}

    @property
    def self(self):
        """
        Dictionary which if passed to its own constructor will yield an identical copy.

        .. note:: This differs from :attr:`pystruct` and :attr:`defining_dict` in that it uses the
                  hidden variable value, if it exists, instead of the exposed one. This prevents
                  from, for example, passing a value which is 10**10**val (and recurring!).
        """
        # Try to first use the hidden variable before using the non-hidden variety.
        dct = {}
        for k in self._defaults_:
            if hasattr(self, "_" + k):
                dct[k] = getattr(self, "_" + k)
            else:
                dct[k] = getattr(self, k)

        return dct

    def __repr__(self):
        """Full unique representation of the instance."""
        return (
            self.__class__.__name__
            + "("
            + ", ".join(sorted(k + ":" + str(v) for k, v in self.defining_dict.items()))
            + ")"
        )

    def __eq__(self, other):
        """Check whether this instance is equal to another object (by checking the __repr__)."""
        return self.__repr__() == repr(other)

    def __hash__(self):
        """Generate a unique hsh for the instance."""
        return hash(self.__repr__())

    def __str__(self):
        """Human-readable string representation of the object."""
        biggest_k = max(len(k) for k in self.defining_dict)
        params = "\n    ".join(
            sorted(f"{k:<{biggest_k}}: {v}" for k, v in self.defining_dict.items())
        )
        return f"""{self.__class__.__name__}:
    {params}
    """


def snake_to_camel(word: str, publicize: bool = True):
    """Convert snake case to camel case."""
    if publicize:
        word = word.lstrip("_")
    return "".join(x.capitalize() or "_" for x in word.split("_"))


def camel_to_snake(word: str, depublicize: bool = False):
    """Convert came case to snake case."""
    word = "".join("_" + i.lower() if i.isupper() else i for i in word)

    if not depublicize:
        word = word.lstrip("_")

    return word


def get_all_subclasses(cls):
    """Get a list of all subclasses of a given class, recursively."""
    all_subclasses = []

    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))

    return all_subclasses


class OutputStruct(StructWrapper, metaclass=ABCMeta):
    """Base class for any class that wraps a C struct meant to be output from a C function."""

    _meta = True
    _fields_ = []
    _global_params = None
    _inputs = ("user_params", "cosmo_params", "_random_seed")
    _filter_params = ["external_table_path", "wisdoms_path"]
    _c_based_pointers = ()
    _c_compute_function = None

    _TYPEMAP = bidict({"float32": "float *", "float64": "double *", "int32": "int *"})

    def __init__(self, *, random_seed=None, dummy=False, initial=False, **kwargs):
        """
        Base type for output structures from C functions.

        Parameters
        ----------
        random_seed
            Seed associated with the output.
        dummy
            Specify this as a dummy struct, in which no arrays are to be
            initialized or computed.
        initial
            Specify this as an initial struct, where arrays are to be
            initialized, but do not need to be computed to pass into another
            struct's compute().
        """
        super().__init__()

        self.version = ".".join(__version__.split(".")[:2])
        self.patch_version = ".".join(__version__.split(".")[2:])
        self._paths = []

        self._random_seed = random_seed

        for k in self._inputs:

            if k not in self.__dict__:
                try:
                    setattr(self, k, kwargs.pop(k))
                except KeyError:
                    raise KeyError(
                        f"{self.__class__.__name__} requires the keyword argument {k}"
                    )

        if kwargs:
            warnings.warn(
                f"{self.__class__.__name__} received the following unexpected "
                f"arguments: {list(kwargs.keys())}"
            )

        self.dummy = dummy
        self.initial = initial

        self._array_structure = self._get_box_structures()
        self._array_state = {k: ArrayState() for k in self._array_structure}
        self._array_state.update({k: ArrayState() for k in self._c_based_pointers})

        for k in self._array_structure:
            if k not in self.pointer_fields:
                raise TypeError(f"Key {k} in {self} not a defined pointer field in C.")

    @property
    def path(self) -> Tuple[None, Path]:
        """The path to an on-disk version of this object."""
        if not self._paths:
            return None

        for pth in self._paths:
            if pth.exists():
                return pth

        logger.info("All paths that defined {self} have been deleted on disk.")
        return None

    @abstractmethod
    def _get_box_structures(self) -> Dict[str, Union[Dict, Tuple[int]]]:
        """Return a dictionary of names mapping to shapes for each array in the struct.

        The reason this is a function, not a simple attribute, is that we may need to
        decide on what arrays need to be initialized based on the inputs (eg. if USE_2LPT
        is True or False).

        Each actual OutputStruct subclass needs to implement this. Note that the arrays
        are not actually initialized here -- that's done automatically by :func:`_init_arrays`
        using this information. This function means that the names of the actually required
        arrays can be accessed without doing any actual initialization.

        Note also that this only contains arrays allocated *by Python* not C. Arrays
        allocated by C are specified in :func:`_c_shape`.
        """
        pass

    def _c_shape(self, cstruct) -> Dict[str, Tuple[int]]:
        """Return a dictionary of field: shape for arrays allocated within C."""
        return {}

    @classmethod
    def _implementations(cls):
        all_classes = get_all_subclasses(cls)
        return [c for c in all_classes if not c._meta]

    def _init_arrays(self):
        for k, state in self._array_state.items():
            if k == "lowres_density":
                logger.debug("THINKING ABOUT INITING LOWRES_DENSITY")
                logger.debug(state.initialized, state.computed_in_mem, state.on_disk)

            # Don't initialize C-based pointers or already-inited stuff, or stuff
            # that's computed on disk (if it's on disk, accessing the array should
            # just give the computed version, which is what we would want, not a
            # zero-inited array).
            if k in self._c_based_pointers or state.initialized or state.on_disk:
                continue

            params = self._array_structure[k]
            tp = self._TYPEMAP.inverse[self.fields[k].type.cname]

            if isinstance(params, tuple):
                shape = params
                fnc = np.zeros
            elif isinstance(params, dict):
                fnc = params.get("init", np.zeros)
                shape = params.get("shape")
            else:
                raise ValueError("params is not a tuple or dict")

            setattr(self, k, fnc(shape, dtype=tp))

            # Add it to initialized arrays.
            state.initialized = True

    @property
    def random_seed(self):
        """The random seed for this particular instance."""
        if self._random_seed is None:
            self._random_seed = int(np.random.randint(1, int(1e12)))

        return self._random_seed

    def _init_cstruct(self):
        # Initialize all uninitialized arrays.
        self._init_arrays()

        for k, state in self._array_state.items():
            # We do *not* set COMPUTED_ON_DISK items to the C-struct here, because we have no
            # way of knowing (in this function) what is required to load in, and we don't want
            # to unnecessarily load things in. We leave it to the user to ensure that all
            # required arrays are loaded into memory before calling this function.
            if state.initialized:
                setattr(self._cstruct, k, self._ary2buf(getattr(self, k)))

        for k in self.primitive_fields:
            try:
                setattr(self._cstruct, k, getattr(self, k))
            except AttributeError:
                pass

    def _ary2buf(self, ary):
        if not isinstance(ary, np.ndarray):
            raise ValueError("ary must be a numpy array")
        return self._ffi.cast(
            OutputStruct._TYPEMAP[ary.dtype.name], self._ffi.from_buffer(ary)
        )

    def __call__(self):
        """Initialize/allocate a fresh C struct in memory and return it."""
        if not self.dummy:
            self._init_cstruct()

        return self._cstruct

    def __expose(self):
        """Expose the non-array primitives of the ctype to the top-level object."""
        for k in self.primitive_fields:
            setattr(self, k, getattr(self._cstruct, k))

    @property
    def _fname_skeleton(self):
        """The filename without specifying the random seed."""
        return self._name + "_" + self._md5 + "_r{seed}.h5"

    def prepare(
        self,
        flush: Optional[Sequence[str]] = None,
        keep: Optional[Sequence[str]] = None,
        force: bool = False,
    ):
        """Prepare the instance for being passed to another function.

        This will flush all arrays in "flush" from memory, and ensure all arrays
        in "keep" are in memory. At least one of these must be provided. By default,
        the complement of the given parameter is all flushed/kept.


        Parameters
        ----------
        flush
            Arrays to flush out of memory. Note that if no file is associated with this
            instance, these arrays will be lost forever.
        keep
            Arrays to keep or load into memory. Note that if these do not already
            exist, they will be loaded from file (if the file exists). Only one of
            ``flush`` and ``keep`` should be specified.
        force
            Whether to force flushing arrays even if no disk storage exists.
        """
        if flush is None and keep is None:
            raise ValueError("Must provide either flush or keep")

        if flush is not None and keep is None:
            keep = [k for k in self._array_state if k not in flush]
        elif keep is not None and flush is None:
            flush = [k for k in self._array_state if k not in keep]

        flush = flush or []
        keep = keep or []

        for k in flush:
            self._remove_array(k, force)

        # Accessing the array loads it into memory.
        for k in keep:
            getattr(self, k)

    def _remove_array(self, k, force=False):
        state = self._array_state[k]

        if not state.initialized:
            warnings.warn(f"Trying to remove array that isn't yet created: {k}")
            return

        if state.computed_in_mem and not state.on_disk and not force:
            raise OSError(
                f"Trying to purge array '{k}' from memory that hasn't been stored! Use force=True if you meant to do this."
            )

        if state.c_has_active_memory:
            lib.free(getattr(self._cstruct, k))

        delattr(self, k)
        state.initialized = False

    def __getattr__(self, item):
        """Gets arrays that aren't already in memory."""
        # Have to use __dict__ here to test membership, otherwise we get recursion error.
        if "_array_state" not in self.__dict__ or item not in self._array_state:
            raise self.__getattribute__(item)

        if not self._array_state[item].on_disk:
            raise OSError(
                f"Cannot get {item} as it is not in memory, and this object is not cached to disk."
            )

        self.read(fname=self.path, keys=[item])
        return getattr(self, item)

    def purge(self, force=False):
        """Flush all the boxes out of memory.

        Parameters
        ----------
        force
            Whether to force the purge even if no disk storage exists.
        """
        self.prepare(keep=[], force=force)

    def load_all(self):
        """Load all possible arrays into memory."""
        self.prepare(flush=[])

    @property
    def filename(self):
        """The base filename of this object."""
        if self._random_seed is None:
            raise AttributeError("filename not defined until random_seed has been set")

        return self._fname_skeleton.format(seed=self.random_seed)

    def _get_fname(self, direc=None):
        direc = path.abspath(path.expanduser(direc or config["direc"]))
        return path.join(direc, self.filename)

    def _find_file_without_seed(self, direc):
        allfiles = glob.glob(path.join(direc, self._fname_skeleton.format(seed="*")))

        if allfiles:
            return allfiles[0]
        else:
            return None

    def find_existing(self, direc=None):
        """
        Try to find existing boxes which match the parameters of this instance.

        Parameters
        ----------
        direc : str, optional
            The directory in which to search for the boxes. By default, this is the
            centrally-managed directory, given by the ``config.yml`` in ``~/.21cmfast/``.

        Returns
        -------
        str
            The filename of an existing set of boxes, or None.
        """
        # First, if appropriate, find a file without specifying seed.
        # Need to do this first, otherwise the seed will be chosen randomly upon
        # choosing a filename!
        direc = path.expanduser(direc or config["direc"])

        if not self._random_seed:
            f = self._find_file_without_seed(direc)
            if f and self._check_parameters(f):
                return f
        else:
            f = self._get_fname(direc)
            if path.exists(f) and self._check_parameters(f):
                return f
        return None

    def _check_parameters(self, fname):
        with h5py.File(fname, "r") as f:
            for k in self._inputs + ("_global_params",):
                q = getattr(self, k)

                # The key name as it should appear in file.
                kfile = k.lstrip("_")

                # If this particular variable is set to None, this is interpreted
                # as meaning that we don't care about matching it to file.
                if q is None:
                    continue

                if (
                    not isinstance(q, StructWithDefaults)
                    and not isinstance(q, StructInstanceWrapper)
                    and f.attrs[kfile] != q
                ):
                    return False
                elif isinstance(q, (StructWithDefaults, StructInstanceWrapper)):
                    grp = f[kfile]

                    dct = q.self if isinstance(q, StructWithDefaults) else q
                    for kk, v in dct.items():
                        if kk not in self._filter_params:
                            file_v = grp.attrs[kk]
                            if file_v == "none":
                                file_v = None
                            if file_v != v:
                                logger.debug("For file %s:" % fname)
                                logger.debug(
                                    f"\tThough md5 and seed matched, the parameter {kk} did not match,"
                                    f" with values {file_v} and {v} in file and user respectively"
                                )
                                return False
        return True

    def exists(self, direc=None):
        """
        Return a bool indicating whether a box matching the parameters of this instance is in cache.

        Parameters
        ----------
        direc : str, optional
            The directory in which to search for the boxes. By default, this is the
            centrally-managed directory, given by the ``config.yml`` in ``~/.21cmfast/``.
        """
        return self.find_existing(direc) is not None

    def write(self, direc=None, fname=None, write_inputs=True, mode="w"):
        """
        Write the struct in standard HDF5 format.

        Parameters
        ----------
        direc : str, optional
            The directory in which to write the boxes. By default, this is the
            centrally-managed directory, given by the ``config.yml`` in ``~/.21cmfast/``.
        fname : str, optional
            The filename to write to. By default creates a unique filename from the hash.
        write_inputs : bool, optional
            Whether to write the inputs to the file. Can be useful to set to False if
            the input file already exists and has parts already written.
        """
        if not all(v.computed for v in self._array_state.values()):
            raise OSError(
                "Not all boxes have been computed (or maybe some have been purged). Cannot write."
                f"Non-computed boxes: {[k for k, v in self._array_state.items() if not v.computed]}"
            )

        if not self._random_seed:
            raise ValueError(
                "Attempting to write when no random seed has been set. "
                "Struct has been 'computed' inconsistently."
            )

        if not write_inputs:
            mode = "a"

        try:
            direc = path.expanduser(direc or config["direc"])

            if not path.exists(direc):
                makedirs(direc)

            fname = fname or self._get_fname(direc)
            if not path.isabs(fname):
                fname = path.abspath(path.join(direc, fname))

            with h5py.File(fname, mode) as f:
                # Save input parameters to the file
                if write_inputs:
                    for k in self._inputs + ("_global_params",):
                        q = getattr(self, k)

                        kfile = k.lstrip("_")

                        if isinstance(q, (StructWithDefaults, StructInstanceWrapper)):
                            grp = f.create_group(kfile)
                            dct = q.self if isinstance(q, StructWithDefaults) else q
                            for kk, v in dct.items():
                                if kk not in self._filter_params:
                                    try:
                                        grp.attrs[kk] = "none" if v is None else v
                                    except TypeError:
                                        raise TypeError(
                                            f"key {kk} with value {v} is not able to be written to HDF5 attrs!"
                                        )
                        else:
                            f.attrs[kfile] = q

                    # Write 21cmFAST version to the file
                    f.attrs["version"] = __version__

                # Save the boxes to the file
                boxes = f.create_group(self._name)

                self.write_data_to_hdf5_group(boxes)

                self._paths.insert(0, Path(fname))

        except OSError as e:
            logger.warning(
                f"When attempting to write {self.__class__.__name__} to file, write failed with the following error. Continuing without caching."
            )

            logger.warning(e)

    def write_data_to_hdf5_group(self, group: h5py.Group):
        """
        Write out this object to a particular HDF5 subgroup.

        Parameters
        ----------
        group
            The HDF5 group into which to write the object.
        """
        # Go through all fields in this struct, and save
        for k, state in self._array_state.items():
            group.create_dataset(k, data=getattr(self, k))
            state.on_disk = True

        for k in self.primitive_fields:
            group.attrs[k] = getattr(self, k)

    def save(self, fname=None, direc="."):
        """Save the box to disk.

        In detail, this just calls write, but changes the default directory to the
        local directory. This is more user-friendly, while :meth:`write` is for
        automatic use under-the-hood.

        Parameters
        ----------
        fname : str, optional
            The filename to write. Can be an absolute or relative path. If relative,
            by default it is relative to the current directory (otherwise relative
            to ``direc``). By default, the filename is auto-generated as unique to
            the set of parameters that go into producing the data.
        direc : str, optional
            The directory into which to write the data. By default the current directory.
            Ignored if ``fname`` is an absolute path.
        """
        # If fname is absolute path, then get direc from it, otherwise assume current dir.
        if path.isabs(fname):
            direc = path.dirname(fname)
            fname = path.basename(fname)

        self.write(direc, fname)

    def _get_path(
        self, direc: Union[str, Path, None] = None, fname: Union[str, Path, None] = None
    ) -> Path:
        if direc is None and fname is None and self.path:
            return self.path

        if fname is None:
            pth = self.find_existing(direc)

            if pth is None:
                raise OSError("No boxes exist for these parameters.")
        else:
            direc = Path(direc or config["direc"]).expanduser()
            fname = Path(fname)
            pth = fname if fname.exists() else direc / fname
        return pth

    def read(
        self,
        direc: Union[str, Path, None] = None,
        fname: Union[str, Path, None] = None,
        keys: Optional[Sequence[str]] = None,
    ):
        """
        Try find and read existing boxes from cache, which match the parameters of this instance.

        Parameters
        ----------
        direc
            The directory in which to search for the boxes. By default, this is the
            centrally-managed directory, given by the ``config.yml`` in ``~/.21cmfast/``.
        fname
            The filename to read. By default, use the filename associated with this
            object.
        keys
            The names of boxes to read in (can be a subset). By default, read everything.
        """
        pth = self._get_path(direc, fname)

        with h5py.File(pth, "r") as f:
            try:
                boxes = f[self._name]
            except KeyError:
                raise OSError(
                    f"While trying to read in {self._name}, the file exists, but does not have the "
                    "correct structure."
                )

            # Set our arrays.
            for k in boxes.keys():
                if keys is None or k in keys:
                    setattr(self, k, boxes[k][...])
                    self._array_state[k].on_disk = True
                    self._array_state[k].computed_in_mem = True
                    setattr(self._cstruct, k, self._ary2buf(getattr(self, k)))

            for k in boxes.attrs.keys():
                if k == "version":
                    version = ".".join(boxes.attrs[k].split(".")[:2])
                    patch = ".".join(boxes.attrs[k].split(".")[2:])

                    if version != ".".join(__version__.split(".")[:2]):
                        # Ensure that the major and minor versions are the same.
                        warnings.warn(
                            f"The file {pth} is out of date (version = {version}.{patch}). "
                            f"Consider using another box and removing it!"
                        )

                    self.version = version
                    self.patch_version = patch

                setattr(self, k, boxes.attrs[k])
                try:
                    setattr(self._cstruct, k, getattr(self, k))
                except AttributeError:
                    pass

            # Need to make sure that the seed is set to the one that's read in.
            seed = f.attrs["random_seed"]
            self._random_seed = seed

        self.__expose()
        self._paths.insert(0, Path(pth))

    @classmethod
    def from_file(cls, fname, direc=None, load_data=True):
        """Create an instance from a file on disk.

        Parameters
        ----------
        fname : str, optional
            Path to the file on disk. May be relative or absolute.
        direc : str, optional
            The directory from which fname is relative to (if it is relative). By
            default, will be the cache directory in config.
        load_data : bool, optional
            Whether to read in the data when creating the instance. If False, a bare
            instance is created with input parameters -- the instance can read data
            with the :func:`read` method.
        """
        direc = path.expanduser(direc or config["direc"])

        if not path.exists(fname):
            fname = path.join(direc, fname)

        self = cls(**cls._read_inputs(fname))

        if load_data:
            self.read(fname=fname)
        return self

    @classmethod
    def _read_inputs(cls, fname):
        input_classes = [c.__name__ for c in StructWithDefaults.__subclasses__()]

        # Read the input parameter dictionaries from file.
        kwargs = {}
        with h5py.File(fname, "r") as fl:
            for k in cls._inputs:
                kfile = k.lstrip("_")
                input_class_name = snake_to_camel(kfile)

                if input_class_name in input_classes:
                    input_class = StructWithDefaults.__subclasses__()[
                        input_classes.index(input_class_name)
                    ]
                    grp = fl[kfile]
                    kwargs[k] = input_class(
                        {k: v for k, v in dict(grp.attrs).items() if v != "none"}
                    )
                else:
                    kwargs[kfile] = fl.attrs[kfile]
        return kwargs

    def __repr__(self):
        """Return a fully unique representation of the instance."""
        # This is the class name and all parameters which belong to C-based input structs,
        # eg. InitialConditions(HII_DIM:100,SIGMA_8:0.8,...)
        # eg. InitialConditions(HII_DIM:100,SIGMA_8:0.8,...)
        return f"{self._seedless_repr()}_random_seed={self._random_seed}"

    def _seedless_repr(self):
        # The same as __repr__ except without the seed.
        return (
            (
                self._name
                + "("
                + "; ".join(
                    repr(v)
                    if isinstance(v, StructWithDefaults)
                    else (
                        v.filtered_repr(self._filter_params)
                        if isinstance(v, StructInstanceWrapper)
                        else k.lstrip("_") + ":" + repr(v)
                    )
                    for k, v in [
                        (k, getattr(self, k))
                        for k in self._inputs + ("_global_params",)
                        if k != "_random_seed"
                    ]
                )
            )
            + f"; v{self.version}"
            + ")"
        )

    def __str__(self):
        """Return a human-readable representation of the instance."""
        # this is *not* a unique representation, and doesn't include global params.
        return (
            self._name
            + "("
            + ";\n\t".join(
                repr(v)
                if isinstance(v, StructWithDefaults)
                else k.lstrip("_") + ":" + repr(v)
                for k, v in [(k, getattr(self, k)) for k in self._inputs]
            )
        ) + ")"

    def __hash__(self):
        """Return a unique hsh for this instance, even global params and random seed."""
        return hash(repr(self))

    @property
    def _md5(self):
        """Return a unique hsh of the object, *not* taking into account the random seed."""
        return md5(self._seedless_repr().encode()).hexdigest()

    def __eq__(self, other):
        """Check equality with another object via its __repr__."""
        return repr(self) == repr(other)

    @property
    def is_computed(self) -> bool:
        """Whether this instance has been computed at all.

        This is true either if the current instance has called :meth:`compute`,
        or if it has a current existing :attr:`path` pointing to stored data,
        or if such a path exists.

        Just because the instance has been computed does *not* mean that all
        relevant quantities are available -- some may have been purged from
        memory without writing. Use :meth:`has` to check whether certain arrays
        are available.
        """
        return any(v.computed for v in self._array_state.values())

    def ensure_arrays_computed(self, *arrays, load=False) -> bool:
        """Check if the given arrays are computed (not just initialized)."""
        if not self.is_computed:
            return False

        computed = all(self._array_state[k].computed for k in arrays)

        if computed and load:
            self.prepare(keep=arrays, flush=[])

        return computed

    def ensure_arrays_inited(self, *arrays, init=False) -> bool:
        """Check if the given arrays are initialized (or computed)."""
        inited = all(self._array_state[k].initialized for k in arrays)

        if init and not inited:
            self._init_arrays()

        return True

    @abstractmethod
    def get_required_input_arrays(self, input_box) -> List[str]:
        """Return all input arrays required to compute this object."""
        pass

    def ensure_input_computed(self, input_box, load=False) -> bool:
        """Ensure all the inputs have been computed."""
        if input_box.dummy:
            return True

        arrays = self.get_required_input_arrays(input_box)

        if input_box.initial:
            return input_box.ensure_arrays_inited(*arrays, init=load)

        return input_box.ensure_arrays_computed(*arrays, load=load)

    def summarize(self, indent=0) -> str:
        """Generate a string summary of the struct."""
        indent = indent * "    "
        out = f"\n{indent}{self.__class__.__name__}\n"

        out += "".join(
            f"{indent}    {fieldname:>15}: {getattr(self, fieldname, 'non-existent')}\n"
            for fieldname in self.primitive_fields
        )

        for fieldname, state in self._array_state.items():
            if not state.initialized:
                out += f"{indent}    {fieldname:>15}: uninitialized\n"
            elif not state.computed:
                out += f"{indent}    {fieldname:>15}: initialized\n"
            elif not state.computed_in_mem:
                out += f"{indent}    {fieldname:>15}: computed on disk\n"
            else:
                x = getattr(self, fieldname).flatten()
                if len(x) > 0:
                    out += f"{indent}    {fieldname:>15}: {x[0]:1.4e}, {x[-1]:1.4e}, {x.min():1.4e}, {x.max():1.4e}, {np.mean(x):1.4e}\n"
                else:
                    out += f"{indent}    {fieldname:>15}: size zero\n"

        return out

    @classmethod
    def _log_call_arguments(cls, *args):
        logger.debug(f"Calling {cls._c_compute_function.__name__} with following args:")

        for arg in args:
            if isinstance(arg, OutputStruct):
                for line in arg.summarize(indent=1).split("\n"):
                    logger.debug(line)
            elif isinstance(arg, StructWithDefaults):
                for line in str(arg).split("\n"):
                    logger.debug(f"    {line}")
            else:
                logger.debug(f"    {arg}")

    def _ensure_arguments_exist(self, *args):
        for arg in args:
            if (
                isinstance(arg, OutputStruct)
                and not arg.dummy
                and not self.ensure_input_computed(arg, load=True)
            ):
                raise ValueError(
                    f"Trying to use {arg} to compute {self}, but some required arrays "
                    f"are not computed!"
                )

    def _compute(
        self, *args, hooks: Optional[Dict[Union[str, Callable], Dict[str, Any]]] = None
    ):
        """Compute the actual function that fills this struct."""
        # Write a detailed message about call arguments if debug turned on.
        if logger.getEffectiveLevel() <= logging.DEBUG:
            self._log_call_arguments(*args)

        # Check that all required inputs are really computed, and load them into memory
        # if they're not already.
        self._ensure_arguments_exist(*args)

        # Construct the args. All StructWrapper objects need to actually pass their
        # underlying cstruct, rather than themselves. OutputStructs also pass the
        # class in that's calling this.
        inputs = [arg() if isinstance(arg, StructWrapper) else arg for arg in args]

        # Ensure we haven't already tried to compute this instance.
        if self.is_computed:
            raise ValueError(
                f"You are trying to compute {self.__class__.__name__}, but it has already been computed."
            )

        # Perform the C computation
        try:
            exitcode = self._c_compute_function(*inputs, self())
        except TypeError as e:
            logger.error(
                f"Arguments to {self._c_compute_function.__name__}: "
                f"{[arg() if isinstance(arg, StructWrapper) else arg for arg in args]}"
            )
            raise e

        _process_exitcode(exitcode, self._c_compute_function, args)

        # Ensure memory created in C gets mapped to numpy arrays in this struct.
        for k, state in self._array_state.items():
            if state.initialized:
                state.computed_in_mem = True

        self.__memory_map()
        self.__expose()

        # Optionally do stuff with the result (like writing it)
        self._call_hooks(hooks)

        return self

    def _call_hooks(self, hooks):
        if hooks is None:
            hooks = {"write": {"direc": config["direc"]}}

        for hook, params in hooks.items():

            if callable(hook):
                hook(self, **params)
            else:
                getattr(self, hook)(**params)

    def __memory_map(self):
        shapes = self._c_shape(self._cstruct)
        for item in self._c_based_pointers:
            setattr(self, item, asarray(getattr(self._cstruct, item), shapes[item]))
            self._array_state[item].c_memory = True
            self._array_state[item].computed_in_mem = True

    def __del__(self):
        """Safely delete the object and its C-allocated memory."""
        for k in self._c_based_pointers:
            if self._array_state[k].c_has_active_memory:
                lib.free(getattr(self._cstruct, k))


class StructInstanceWrapper:
    """A wrapper for *instances* of C structs.

    This is as opposed to :class:`StructWrapper`, which is for the un-instantiated structs.

    Parameters
    ----------
    wrapped :
        The reference to the C object to wrap (contained in the ``cffi.lib`` object).
    ffi :
        The ``cffi.ffi`` object.
    """

    def __init__(self, wrapped, ffi):
        self._cobj = wrapped
        self._ffi = ffi

        for nm, tp in self._ffi.typeof(self._cobj).fields:
            setattr(self, nm, getattr(self._cobj, nm))

        # Get the name of the structure
        self._ctype = self._ffi.typeof(self._cobj).cname.split()[-1]

    def __setattr__(self, name, value):
        """Set an attribute of the instance, attempting to change it in the C struct as well."""
        try:
            setattr(self._cobj, name, value)
        except AttributeError:
            pass
        object.__setattr__(self, name, value)

    def items(self):
        """Yield (name, value) pairs for each element of the struct."""
        for nm, tp in self._ffi.typeof(self._cobj).fields:
            yield nm, getattr(self, nm)

    def keys(self):
        """Return a list of names of elements in the struct."""
        return [nm for nm, tp in self.items()]

    def __repr__(self):
        """Return a unique representation of the instance."""
        return (
            self._ctype
            + "("
            + ";".join(k + "=" + str(v) for k, v in sorted(self.items()))
        ) + ")"

    def filtered_repr(self, filter_params):
        """Get a fully unique representation of the instance that filters out some parameters.

        Parameters
        ----------
        filter_params : list of str
            The parameter names which should not appear in the representation.
        """
        return (
            self._ctype
            + "("
            + ";".join(
                k + "=" + str(v)
                for k, v in sorted(self.items())
                if k not in filter_params
            )
        ) + ")"


def _check_compatible_inputs(*datasets, ignore=["redshift"]):
    """Ensure that all defined input parameters for the provided datasets are equal.

    Parameters
    ----------
    datasets : list of :class:`~_utils.OutputStruct`
        A number of output datasets to cross-check.
    ignore : list of str
        Attributes to ignore when ensuring that parameter inputs are the same.

    Raises
    ------
    ValueError :
        If datasets are not compatible.
    """
    done = []  # keeps track of inputs we've checked so we don't double check.

    for i, d in enumerate(datasets):
        # If a dataset is None, just ignore and move on.
        if d is None:
            continue

        # noinspection PyProtectedMember
        for inp in d._inputs:
            # Skip inputs that we want to ignore
            if inp in ignore:
                continue

            if inp not in done:
                for j, d2 in enumerate(datasets[(i + 1) :]):
                    if d2 is None:
                        continue

                    # noinspection PyProtectedMember
                    if inp in d2._inputs and getattr(d, inp) != getattr(d2, inp):
                        raise ValueError(
                            f"""
                            {d.__class__.__name__} and {d2.__class__.__name__} are incompatible.
                            {inp}: {getattr(d, inp)}
                            vs.
                            {inp}: {getattr(d2, inp)}
                            """
                        )
                done += [inp]



import py21cmfast as p21c


def test_first_box(default_user_params):
    """Tests whether the first_box idea works for spin_temp.
    This test was breaking before we set the z_heat_max box to actually get
    the correct dimensions (before it was treated as a dummy).
    """
    initial_conditions = p21c.initial_conditions(
        user_params=default_user_params.clone(HII_DIM=default_user_params.HII_DIM + 1),
        random_seed=1,
    )

    perturbed_field = p21c.perturb_field(redshift=29.0, init_boxes=initial_conditions)

    spin_temp = p21c.spin_temperature(perturbed_field=perturbed_field, z_heat_max=30.0)

    assert spin_temp.redshift == 29.0



"""Given a log file from Github Actions, takes away stuff that *should* change between runs.

Useful for preparing it to be diff'ed against another log to check where things went wrong.
"""


import sys

fname = sys.argv[1]

if len(sys.argv) > 2:
    ignores = sys.argv[2:]
else:
    ignores = []

pids = set()
threads = set()

pid_map = {}
with open(fname) as fl:

    lines = fl.readlines()

    # Get to the start of the testing
    for i, line in enumerate(lines):
        if "==== test session starts ====" in line:
            break

    print(f"Starting on line {i}")

    for line in lines[i:]:
        pids.add(line.split("pid=")[1].split("/")[0] if "pid=" in line else "other")
        threads.add(line.split("thr=")[1].split("]")[0] if "thr=" in line else "other")

    for ii, pid in enumerate(sorted(pids)):
        pid_map[pid] = ii

    out = {p: {t: [] for t in threads} for p in pid_map}

    for line in lines[(i + 1) :]:
        if "======" in line:
            break

        if any(ign in line for ign in ignores):
            continue

        ln = "|".join(line.split("|")[1:])
        pid = ln.split("pid=")[1].split("/")[0] if "pid=" in ln else "other"
        thread = ln.split("thr=")[1].split("]")[0] if "thr=" in ln else "other"

        ln = ln.replace(f"pid={pid}", f"pid={pid_map[pid]}")

        out[pid][thread].append(ln)


with open(fname + ".processed", "w") as fl:
    for key in sorted(pids):
        for t in sorted(threads):
            fl.writelines(out[key][t])



#!/usr/bin/python
import sys
from datetime import datetime

if __name__ == "__main__":
    newversion = sys.argv[1]

    with open("CHANGELOG.rst", "r") as fl:
        lines = fl.readlines()

    for i, line in enumerate(lines):
        if line == "dev-version\n":
            break
    else:
        raise IOError("Couldn't Find 'dev-version' tag")

    lines.insert(i + 2, "----------------------\n")
    lines.insert(i + 2, f'v{newversion} [{datetime.now().strftime("%d %b %Y")}]\n')
    lines.insert(i + 2, "\n")

    with open("CHANGELOG.rst", "w") as fl:
        fl.writelines(lines)



"""
These are designed to be unit-tests of the wrapper functionality. They do not test for
correctness of simulations,
but whether different parameter options work/don't work as intended.
"""
import pytest

import numpy as np

from py21cmfast import wrapper


@pytest.fixture(scope="module")
def perturb_field_lowz(ic, low_redshift):
    """A default perturb_field"""
    return wrapper.perturb_field(redshift=low_redshift, init_boxes=ic, write=True)


@pytest.fixture(scope="module")
def ionize_box(perturb_field):
    """A default ionize_box"""
    return wrapper.ionize_box(perturbed_field=perturb_field, write=True)


@pytest.fixture(scope="module")
def ionize_box_lowz(perturb_field_lowz):
    """A default ionize_box at lower redshift."""
    return wrapper.ionize_box(perturbed_field=perturb_field_lowz, write=True)


@pytest.fixture(scope="module")
def spin_temp(perturb_field):
    """A default perturb_field"""
    return wrapper.spin_temperature(perturbed_field=perturb_field, write=True)


def test_perturb_field_no_ic(default_user_params, redshift, perturb_field):
    """Run a perturb field without passing an init box"""
    pf = wrapper.perturb_field(redshift=redshift, user_params=default_user_params)
    assert len(pf.density) == pf.user_params.HII_DIM == default_user_params.HII_DIM
    assert pf.redshift == redshift
    assert pf.random_seed != perturb_field.random_seed
    assert not np.all(pf.density == 0)
    assert pf != perturb_field
    assert pf._seedless_repr() == perturb_field._seedless_repr()


def test_ib_no_z(ic):
    with pytest.raises(ValueError):
        wrapper.ionize_box(init_boxes=ic)


def test_pf_unnamed_param():
    """Try using an un-named parameter."""
    with pytest.raises(TypeError):
        wrapper.perturb_field(7)


def test_perturb_field_ic(perturb_field, ic):
    # this will run perturb_field again, since by default regenerate=True for tests.
    # BUT it should produce exactly the same as the default perturb_field since it has
    # the same seed.
    pf = wrapper.perturb_field(redshift=perturb_field.redshift, init_boxes=ic)

    assert len(pf.density) == len(ic.lowres_density)
    assert pf.cosmo_params == ic.cosmo_params
    assert pf.user_params == ic.user_params
    assert not np.all(pf.density == 0)

    assert pf.user_params == perturb_field.user_params
    assert pf.cosmo_params == perturb_field.cosmo_params

    assert pf == perturb_field


def test_cache_exists(default_user_params, perturb_field, tmpdirec):
    pf = wrapper.PerturbedField(
        redshift=perturb_field.redshift,
        cosmo_params=perturb_field.cosmo_params,
        user_params=default_user_params,
    )

    assert pf.exists(tmpdirec)

    pf.read(tmpdirec)
    assert np.all(pf.density == perturb_field.density)
    assert pf == perturb_field


def test_pf_new_seed(perturb_field, tmpdirec):
    pf = wrapper.perturb_field(
        redshift=perturb_field.redshift,
        user_params=perturb_field.user_params,
        random_seed=1,
    )

    # we didn't write it, and this has a different seed
    assert not pf.exists(direc=tmpdirec)
    assert pf.random_seed != perturb_field.random_seed

    assert not np.all(pf.density == perturb_field.density)


def test_ib_new_seed(ionize_box_lowz, perturb_field_lowz, tmpdirec):
    # this should fail because perturb_field has a seed set already, which isn't 1.
    with pytest.raises(ValueError):
        wrapper.ionize_box(
            perturbed_field=perturb_field_lowz,
            random_seed=1,
        )

    ib = wrapper.ionize_box(
        cosmo_params=perturb_field_lowz.cosmo_params,
        redshift=perturb_field_lowz.redshift,
        user_params=perturb_field_lowz.user_params,
        random_seed=1,
    )

    # we didn't write it, and this has a different seed
    assert not ib.exists(direc=tmpdirec)
    assert ib.random_seed != ionize_box_lowz.random_seed
    assert not np.all(ib.xH_box == ionize_box_lowz.xH_box)


def test_st_new_seed(spin_temp, perturb_field, tmpdirec):
    # this should fail because perturb_field has a seed set already, which isn't 1.
    with pytest.raises(ValueError):
        wrapper.spin_temperature(
            perturbed_field=perturb_field,
            random_seed=1,
        )

    st = wrapper.spin_temperature(
        cosmo_params=spin_temp.cosmo_params,
        user_params=spin_temp.user_params,
        astro_params=spin_temp.astro_params,
        flag_options=spin_temp.flag_options,
        redshift=spin_temp.redshift,
        random_seed=1,
    )

    # we didn't write it, and this has a different seed
    assert not st.exists(direc=tmpdirec)
    assert st.random_seed != spin_temp.random_seed
    assert not np.all(st.Ts_box == spin_temp.Ts_box)


def test_st_from_z(perturb_field_lowz, spin_temp):

    # This one has all the same parameters as the nominal spin_temp, but is evaluated
    # with an interpolated perturb_field
    st = wrapper.spin_temperature(
        perturbed_field=perturb_field_lowz,
        astro_params=spin_temp.astro_params,
        flag_options=spin_temp.flag_options,
        redshift=spin_temp.redshift,  # Higher redshift
    )

    assert st != spin_temp
    assert not np.all(st.Ts_box == spin_temp.Ts_box)


def test_ib_from_pf(perturb_field):
    ib = wrapper.ionize_box(perturbed_field=perturb_field)
    assert ib.redshift == perturb_field.redshift
    assert ib.user_params == perturb_field.user_params
    assert ib.cosmo_params == perturb_field.cosmo_params


def test_ib_from_z(default_user_params, perturb_field):
    ib = wrapper.ionize_box(
        redshift=perturb_field.redshift,
        user_params=default_user_params,
        regenerate=False,
    )
    assert ib.redshift == perturb_field.redshift
    assert ib.user_params == perturb_field.user_params
    assert ib.cosmo_params == perturb_field.cosmo_params
    assert ib.cosmo_params is not perturb_field.cosmo_params


def test_ib_override_z(perturb_field):
    with pytest.raises(ValueError):
        wrapper.ionize_box(
            redshift=perturb_field.redshift + 1,
            perturbed_field=perturb_field,
        )


def test_ib_override_z_heat_max(perturb_field):
    # save previous z_heat_max
    zheatmax = wrapper.global_params.Z_HEAT_MAX

    wrapper.ionize_box(
        redshift=perturb_field.redshift,
        perturbed_field=perturb_field,
        z_heat_max=12.0,
    )

    assert wrapper.global_params.Z_HEAT_MAX == zheatmax


def test_ib_bad_st(ic, redshift):
    with pytest.raises(ValueError):
        wrapper.ionize_box(redshift=redshift, spin_temp=ic)


def test_bt(ionize_box, spin_temp, perturb_field):
    with pytest.raises(TypeError):  # have to specify param names
        wrapper.brightness_temperature(ionize_box, spin_temp, perturb_field)

    # this will fail because ionized_box was not created with spin temperature.
    with pytest.raises(ValueError):
        wrapper.brightness_temperature(
            ionized_box=ionize_box, perturbed_field=perturb_field, spin_temp=spin_temp
        )

    bt = wrapper.brightness_temperature(
        ionized_box=ionize_box, perturbed_field=perturb_field
    )

    assert bt.cosmo_params == perturb_field.cosmo_params
    assert bt.user_params == perturb_field.user_params
    assert bt.flag_options == ionize_box.flag_options
    assert bt.astro_params == ionize_box.astro_params


def test_coeval_against_direct(ic, perturb_field, ionize_box):
    coeval = wrapper.run_coeval(perturb=perturb_field, init_box=ic)

    assert coeval.init_struct == ic
    assert coeval.perturb_struct == perturb_field
    assert coeval.ionization_struct == ionize_box


def test_lightcone(lc, default_user_params, redshift, max_redshift):
    assert lc.lightcone_redshifts[-1] >= max_redshift
    assert np.isclose(lc.lightcone_redshifts[0], redshift, atol=1e-4)
    assert lc.cell_size == default_user_params.BOX_LEN / default_user_params.HII_DIM


def test_lightcone_quantities(ic, max_redshift, perturb_field):
    lc = wrapper.run_lightcone(
        init_box=ic,
        perturb=perturb_field,
        max_redshift=max_redshift,
        lightcone_quantities=("dNrec_box", "density", "brightness_temp"),
        global_quantities=("density", "Gamma12_box"),
    )

    assert hasattr(lc, "dNrec_box")
    assert hasattr(lc, "density")
    assert hasattr(lc, "global_density")
    assert hasattr(lc, "global_Gamma12")

    # dNrec is not filled because we're not doing INHOMO_RECO
    assert lc.dNrec_box.max() == lc.dNrec_box.min() == 0

    # density should be filled with not zeros.
    assert lc.density.min() != lc.density.max() != 0

    # Simply ensure that different quantities are not getting crossed/referred to each other.
    assert lc.density.min() != lc.brightness_temp.min() != lc.brightness_temp.max()

    # Raise an error since we're not doing spin temp.
    with pytest.raises(ValueError):
        wrapper.run_lightcone(
            init_box=ic,
            perturb=perturb_field,
            max_redshift=20.0,
            lightcone_quantities=("Ts_box", "density"),
        )

    # And also raise an error for global quantities.
    with pytest.raises(ValueError):
        wrapper.run_lightcone(
            init_box=ic,
            perturb=perturb_field,
            max_redshift=20.0,
            global_quantities=("Ts_box",),
        )


def test_run_lf():
    muv, mhalo, lf = wrapper.compute_luminosity_function(redshifts=[7, 8, 9], nbins=100)
    assert np.all(lf[~np.isnan(lf)] > -30)
    assert lf.shape == (3, 100)

    # Check that memory is in-tact and a second run also works:
    muv, mhalo, lf2 = wrapper.compute_luminosity_function(
        redshifts=[7, 8, 9], nbins=100
    )
    assert lf2.shape == (3, 100)
    assert np.allclose(lf2[~np.isnan(lf2)], lf[~np.isnan(lf)])

    muv_minih, mhalo_minih, lf_minih = wrapper.compute_luminosity_function(
        redshifts=[7, 8, 9],
        nbins=100,
        component=0,
        flag_options={"USE_MINI_HALOS": True},
        mturnovers=[7.0, 7.0, 7.0],
        mturnovers_mini=[5.0, 5.0, 5.0],
    )
    assert np.all(lf_minih[~np.isnan(lf_minih)] > -30)
    assert lf_minih.shape == (3, 100)


def test_coeval_st(ic, perturb_field):
    coeval = wrapper.run_coeval(
        init_box=ic,
        perturb=perturb_field,
        flag_options={"USE_TS_FLUCT": True},
    )

    assert isinstance(coeval.spin_temp_struct, wrapper.TsBox)


def _global_Tb(coeval_box):
    assert isinstance(coeval_box, wrapper.Coeval)
    global_Tb = coeval_box.brightness_temp.mean(dtype=np.float128).astype(np.float32)
    assert np.isclose(global_Tb, coeval_box.brightness_temp_struct.global_Tb)
    return global_Tb


def test_coeval_callback(ic, max_redshift, perturb_field):
    lc, coeval_output = wrapper.run_lightcone(
        init_box=ic,
        perturb=perturb_field,
        max_redshift=max_redshift,
        lightcone_quantities=("brightness_temp",),
        global_quantities=("brightness_temp",),
        coeval_callback=_global_Tb,
    )
    assert isinstance(lc, wrapper.LightCone)
    assert isinstance(coeval_output, list)
    assert len(lc.node_redshifts) == len(coeval_output)
    assert np.allclose(
        lc.global_brightness_temp, np.array(coeval_output, dtype=np.float32)
    )


def test_coeval_callback_redshifts(ic, redshift, max_redshift, perturb_field):
    coeval_callback_redshifts = np.array(
        [max_redshift, max_redshift, (redshift + max_redshift) / 2, redshift],
        dtype=np.float32,
    )
    lc, coeval_output = wrapper.run_lightcone(
        init_box=ic,
        perturb=perturb_field,
        max_redshift=max_redshift,
        coeval_callback=lambda x: x.redshift,
        coeval_callback_redshifts=coeval_callback_redshifts,
    )
    assert len(coeval_callback_redshifts) - 1 == len(coeval_output)
    computed_redshifts = [
        lc.node_redshifts[np.argmin(np.abs(i - lc.node_redshifts))]
        for i in coeval_callback_redshifts[1:]
    ]
    assert np.allclose(coeval_output, computed_redshifts)


def Heaviside(x):
    return 1 if x > 0 else 0


def test_coeval_callback_exceptions(ic, redshift, max_redshift, perturb_field):
    # should output warning in logs and not raise an error
    lc, coeval_output = wrapper.run_lightcone(
        init_box=ic,
        perturb=perturb_field,
        max_redshift=max_redshift,
        coeval_callback=lambda x: 1
        / Heaviside(x.redshift - (redshift + max_redshift) / 2),
        coeval_callback_redshifts=[max_redshift, redshift],
    )
    # should raise an error
    with pytest.raises(RuntimeError) as excinfo:
        lc, coeval_output = wrapper.run_lightcone(
            init_box=ic,
            perturb=perturb_field,
            max_redshift=max_redshift,
            coeval_callback=lambda x: 1 / 0,
            coeval_callback_redshifts=[max_redshift, redshift],
        )
    assert "coeval_callback computation failed on first trial" in str(excinfo.value)


def test_coeval_vs_low_level(ic):

    coeval = wrapper.run_coeval(
        redshift=20,
        init_box=ic,
        zprime_step_factor=1.1,
        regenerate=True,
        flag_options={"USE_TS_FLUCT": True},
        write=False,
    )

    st = wrapper.spin_temperature(
        redshift=20,
        init_boxes=ic,
        zprime_step_factor=1.1,
        regenerate=True,
        flag_options={"USE_TS_FLUCT": True},
        write=False,
    )

    assert np.allclose(coeval.Tk_box, st.Tk_box)
    assert np.allclose(coeval.Ts_box, st.Ts_box)
    assert np.allclose(coeval.x_e_box, st.x_e_box)


def test_using_cached_halo_field(ic, test_direc):
    """Test whether the C-based memory in halo fields is cached correctly.

    Prior to v3.1 this was segfaulting, so this test ensure that this behaviour does
    not regress.
    """
    halo_field = wrapper.determine_halo_list(
        redshift=10.0,
        init_boxes=ic,
        write=True,
        direc=test_direc,
    )

    pt_halos = wrapper.perturb_halo_list(
        redshift=10.0,
        init_boxes=ic,
        halo_field=halo_field,
        write=True,
        direc=test_direc,
    )

    print("DONE WITH FIRST BOXES!")
    # Now get the halo field again at the same redshift -- should be cached
    new_halo_field = wrapper.determine_halo_list(
        redshift=10.0, init_boxes=ic, write=False, regenerate=False
    )

    new_pt_halos = wrapper.perturb_halo_list(
        redshift=10.0,
        init_boxes=ic,
        halo_field=new_halo_field,
        write=False,
        regenerate=False,
    )

    np.testing.assert_allclose(new_halo_field.halo_masses, halo_field.halo_masses)
    np.testing.assert_allclose(pt_halos.halo_coords, new_pt_halos.halo_coords)



#!/usr/bin/env python
"""Setup the package."""


from setuptools import find_packages, setup

import glob
import io
import os
import re
import shutil
from os.path import dirname, expanduser, join


def _read(*names, **kwargs):
    return open(
        join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")
    ).read()


pkgdir = os.path.dirname(os.path.abspath(__file__))


# Enable code coverage for C code: we can't use CFLAGS=-coverage in tox.ini, since that
# may mess with compiling dependencies (e.g. numpy). Therefore we set SETUPPY_
# CFLAGS=-coverage in tox.ini and copy it to CFLAGS here (after deps have been safely installed).
if "TOXENV" in os.environ and "SETUPPY_CFLAGS" in os.environ:
    os.environ["CFLAGS"] = os.environ["SETUPPY_CFLAGS"]

test_req = [
    "pre-commit",
    "pytest>=5.0",
    "pytest-cov",
    "tox",
    "pytest-remotedata>=0.3.2",
    "powerbox",
    "pytest-plt",
    "questionary",
]

doc_req = ["nbsphinx", "numpydoc", "sphinx >= 1.3", "sphinx-rtd-theme"]

setup(
    name="21cmFAST",
    license="MIT license",
    description="A semi-numerical cosmological simulation code for the 21cm signal",
    long_description="%s\n%s"
    % (
        re.compile("^.. start-badges.*^.. end-badges", re.M | re.S).sub(
            "", _read("README.rst")
        ),
        re.sub(":[a-z]+:`~?(.*?)`", r"``\1``", _read("CHANGELOG.rst")),
    ),
    author="The 21cmFAST coredev team",
    author_email="21cmfast.coredev@gmail.com",
    url="https://github.com/21cmFAST/21cmFAST",
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    keywords=["Epoch of Reionization", "Cosmology"],
    install_requires=[
        "click",
        "numpy",
        "pyyaml",
        "cffi>=1.0",
        "scipy",
        "astropy>=2.0",
        "h5py>=2.8.0",
        "cached_property",
        "matplotlib",
        "bidict",
    ],
    extras_require={"tests": test_req, "docs": doc_req, "dev": test_req + doc_req},
    setup_requires=["cffi>=1.0", "setuptools_scm"],
    entry_points={"console_scripts": ["21cmfast = py21cmfast.cli:main"]},
    cffi_modules=[f"{pkgdir}/build_cffi.py:ffi"],
    use_scm_version={"write_to": "src/py21cmfast/_version.py"},
)



"""Run the example in the JOSS paper and make plots."""
import matplotlib.pyplot as plt

import py21cmfast as p21c

lightcone = p21c.run_lightcone(
    redshift=6.0,  # Minimum redshift of the lightcone
    max_redshift=30.0,
    user_params={
        "HII_DIM": 150,  # Number of cells along a side in the output cube
        "DIM": 450,  # Original high-resolution cell number
        "BOX_LEN": 300,  # Size of the simulation in Mpc
        "N_THREADS": 16,
        "USE_INTERPOLATION_TABLES": True,
    },
    flag_options={
        "USE_TS_FLUCT": True,  # Do not assume saturated spin temperature
        "INHOMO_RECO": True,  # Use inhomogeneous recominations
    },
    lightcone_quantities=(  # Component fields to store as interpolated lightcones
        "brightness_temp",
        "xH_box",
        "density",
    ),
    global_quantities=(  # Component fields to store as mean values per redshift
        "xH_box",
        "brightness_temp",
    ),
)

# Save to a unique filename hashing all input parameters
lightcone.save()

fig, ax = plt.subplots(
    1,
    1,
    figsize=(
        7.75,
        7.75 * min(lightcone.shape) / max(lightcone.shape) + 2.5,
    ),
)

# Make a lightcone sliceplot
fig, ax = p21c.plotting.lightcone_sliceplot(
    lightcone, "brightness_temp", fig=fig, ax=ax
)
fig.savefig("lightcone.pdf")

fig, ax = p21c.plotting.plot_global_history(lightcone, "xH")
fig.savefig("xH_history.pdf")
